import itertools
import warnings
from collections.abc import Sequence
from itertools import product
from typing import Optional

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts import concatenate as darts_concatenate
from darts.logging import get_logger, raise_if_not, raise_log
from darts.utils.data.tabularization import create_lagged_prediction_data
from darts.utils.timeseries_generation import linear_timeseries


class TestCreateLaggedPredictionData:
    """
    Tests `create_lagged_prediction_data` function defined in `darts.utils.data.tabularization`. There
    are broadly two 'groups' of tests defined in this module:
        1. 'Generated Test Cases': these test that `create_lagged_prediction_data` produces the same outputs
        as a simplified implementation of the 'time intersection' feature generation method (see
        `darts.utils.data.tabularization` for further details). For these tests, the 'correct answer' is not
        directly specified; instead, it is generated from a set of input parameters using the set of simplified
        functions. The rationale behind this approach is that it allows for many different combinations of input
        values to be effortlessly tested. The drawback of this, however, is that the correctness of these tests
        assumes that the simplified functions have been implemented correctly - if this isn't the case, then these
        tests are not to be trusted. In saying this, these simplified functions are significantly easier to
        understand and debug than the `create_lagged_prediction_data` function they're helping to test.
        2. 'Specified Test Cases': these test that `create_lagged_prediction_data` returns an exactly specified
        output; these specified outputs are *not* 'generated' by another function. Although these 'specified'
        test cases tend to be simpler and less extensive than the 'generated' test cases, their correctness does
        not assume the correct implementation of any other function.
    """

    #
    #   Helper Functions for Generated Test Cases
    #

    @staticmethod
    def create_multivariate_linear_timeseries(
        n_components: int, **kwargs
    ) -> TimeSeries:
        """
        Helper function that creates a `linear_timeseries` with a specified number of
        components. To help distinguish each component from one another, `i` is added on
        to each value of the `i`th component. Any additional key word arguments are passed
        to `linear_timeseries` (`start_value`, `end_value`, `start`, `end`, `length`, etc).
        """
        timeseries = []
        for i in range(n_components):
            # Values of each component is 1 larger than the last:
            timeseries_i = linear_timeseries(**kwargs) + i
            timeseries.append(timeseries_i)
        return darts_concatenate(timeseries, axis=1)

    @staticmethod
    def get_feature_times(
        target: TimeSeries,
        past: TimeSeries,
        future: TimeSeries,
        lags: Optional[int],
        lags_past: Optional[int],
        lags_future: Optional[int],
        max_samples_per_ts: Optional[int],
    ) -> pd.Index:
        """
        Helper function that returns the times shared by all of the specified series that can be used
        to create features and labels. This is performed by using the helper functions
        `get_feature_times_target_or_past` and `get_feature_times_future` (both defined below) to extract
        the feature times from the target series, past covariates, and future covariates respectively, and
        then intersecting these features times with one another. A series is considered to be 'specified'
        if its corresponding lag (e.g. `lags` for `target`, or `lags_future` for `future`) is not `None`.
        If requested, the last `max_samples_per_ts` times are taken.

        This function is basically a simplified implementation of `get_feature_times` in `tabularization.py`
        that only works for `is_training = False`.
        """
        # `times` is initialized inside loop:
        times = None
        all_series = [target, past, future]
        all_lags = [lags, lags_past, lags_future]
        for i, (series_i, lags_i) in enumerate(zip(all_series, all_lags)):
            # If `lags` not specified, ignore series:
            lags_specified = lags_i is not None
            is_target_or_past = i < 2
            if lags_specified:
                if is_target_or_past:
                    times_i = (
                        TestCreateLaggedPredictionData.get_feature_times_target_or_past(
                            series_i, lags_i
                        )
                    )
                else:
                    times_i = TestCreateLaggedPredictionData.get_feature_times_future(
                        series_i, lags_i
                    )
            else:
                times_i = None
            if times_i is not None:
                # Initialise `times` if this is first specified series:
                if times is None:
                    times = times_i
                # Intersect `times` with `times_i` if `times` already initialized:
                else:
                    times = times.intersection(times_i)
        # Take most recent `max_samples_per_ts` samples if requested:
        if (max_samples_per_ts is not None) and (len(times) > max_samples_per_ts):
            times = times[-max_samples_per_ts:]
        return times

    @staticmethod
    def get_feature_times_target_or_past(
        series: TimeSeries,
        lags: Sequence[int],
    ) -> pd.Index:
        """
        Helper function called by `get_feature_times` that extracts all of the times within `target_series`
        *or* `past_covariates` that can be used to create features; because we can assume all of the values
        in `lags` are negative for `target_series` and `past_covariates`, the feature time extract process is
        the same for these series when constructing prediction data.

        More specifically, we can create features for times within `serues` that have at least `max_lag = -min(lags)`
        values preceeding them, since these preceeding values are required to construct a feature vector for that time.
        Since the first `max_lag` times do not fulfill this condition, they are exluded.

        Because we're create features for predicting here, we don't need to worry about whether we can produce
        labels corresponding each time.

        Importantly, features can be constructed for times that occur after the end of `series`: this is because:
            1. We don't need to have all the `series` values up to time `t` to construct a feature for this time;
            instead, we only need to have the values from time `t - min_lag` to `t - max_lag`, where
            `min_lag = -max(lags)` and `max_lag = -min(lags)`. In other words, the latest feature we can create
            for `series` occurs at `series.end_time() + min_lag * series.freq`.
            2. We don't need to use the values of `series` to construct labels when we're creating prediction data,
            so we're able to create a feature for time `t` without having to worry about whether we can construct
            a corresponding label for this time.
        """
        times = series.time_index
        min_lag = -max(lags)
        times = times.union([
            times[-1] + i * series.freq for i in range(1, min_lag + 1)
        ])
        max_lag = -min(lags)
        times = times[max_lag:]
        return times

    @staticmethod
    def get_feature_times_future(
        future_covariates: TimeSeries,
        future_covariates_lags: Sequence[int],
    ) -> pd.Index:
        """
        Helper function called by `get_feature_times` that extracts all of the times within
        `future_covariates` that can be used to create features.

        Unlike the lag values for `target_series` and `past_covariates`, the values in
        `future_covariates_lags` can be negative, zero, or positive. This means that
        `min_lag = -max(future_covariates_lags)` and `max_lag = -min(future_covariates_lags)`
        are *not* guaranteed to be positive here: they could be negative (corresponding to
        a positive value in `future_covariates_lags`), zero, or positive (corresponding to
        a negative value in `future_covariates_lags`). With that being said, the relationship
        `min_lag <= max_lag` always holds.

        Consequently, we need to consider three scenarios when finding feature times
        for `future_covariates`:
            1. Both `min_lag` and `max_lag` are positive, which indicates that all of
            the lag values in `future_covariates_lags` are negative (i.e. only values before
            time `t` are used to create a feature from time `t`). In this case, `min_lag`
            and `max_lag` correspond to the smallest magnitude and largest magnitude *negative*
            lags in `future_covariates_lags` respectively. This means we *can* create features for
            times that extend beyond the end of `future_covariates`; additionally, we're unable
            to create features for the first `min_lag` times (see docstring for `get_feature_times_past`).
            2. Both `min_lag` and `max_lag` are non-positive. In this case, `abs(min_lag)` and `abs(max_lag)`
            correspond to the largest and smallest magnitude lags in `future_covariates_lags` respectively;
            note that, somewhat confusingly, `abs(max_lag) <= abs(min_lag)` here. This means that we *can* create f
            features for times that occur before the start of `future_covariates`; the reasoning for this is
            basically the inverse of Case 1 (i.e. we only need to know the values from times `t + abs(max_lag)`
            to `t + abs(min_lag)` to create a feature for time `t`). Additionally, we're unable to create features
            for the last `abs(min_lag)` times in the series, since these times do not have `abs(min_lag)` values
            after them.
            3. `min_lag` is non-positive (i.e. zero or negative), but `max_lag` is positive. In this case,
            `abs(min_lag)` is the magnitude of the largest *non-negative* lag value in `future_covariates_lags`
            and `max_lag` is the largest *negative* lag value in `future_covariates_lags`. This means that we
            *cannot* create features for times that occur before the start of `future_covariates`, nor for
            times that occur after the end of `future_covariates`; this is because we must have access to
            both times before *and* after time `t` to create a feature for this time, which clearly can't
            be acieved for times extending before the start or after the end of the series. Moreover,
            we must exclude the first `max_lag` times and the last `abs(min_lag)` times, since these
            times do not have enough values before or after them respectively.
        """
        times = future_covariates.time_index
        min_lag = -max(future_covariates_lags)
        max_lag = -min(future_covariates_lags)
        # Case 1:
        if (min_lag > 0) and (max_lag > 0):
            # Can create features for times extending after the end of `future_covariates`:
            times = times.union([
                times[-1] + i * future_covariates.freq for i in range(1, min_lag + 1)
            ])
            # Can't create features for first `max_lag` times in series:
            times = times[max_lag:]
        # Case 2:
        elif (min_lag <= 0) and (max_lag <= 0):
            # Can create features for times before the start of `future_covariates`:
            times = times.union([
                times[0] - i * future_covariates.freq
                for i in range(1, abs(max_lag) + 1)
            ])
            # Can't create features for last `abs(min_lag)` times in series:
            times = times[:min_lag] if min_lag != 0 else times
        # Case 3:
        elif (min_lag <= 0) and (max_lag > 0):
            # Can't create features for last `abs(min_lag)` times in series:
            times = times[:min_lag] if min_lag != 0 else times
            # Can't create features for first `max_lag` times in series:
            times = times[max_lag:]
        # Unexpected case:
        else:
            error_msg = (
                "Unexpected `future_covariates_lags` case encountered: "
                "`min_lag` is positive, but `max_lag` is negative. "
                f"Caused by `future_covariates_lags = {future_covariates_lags}`."
            )
            error = ValueError(error_msg)
            raise_log(error, get_logger(__name__))
        return times

    @staticmethod
    def construct_X_block(
        series: TimeSeries, feature_times: pd.Index, lags: Optional[Sequence[int]]
    ) -> np.array:
        """
        Helper function that creates the lagged features 'block' of a specific
        `series` (i.e. either `target_series`, `past_covariates`, or `future_covariates`);
        the feature matrix `X` is formed by concatenating the blocks of all of the specified
        series along the components axis. If `lags` is `None`, then `None` will be returned in
        lieu of an array. Please refer to the `create_lagged_features` docstring for further
        details about the structure of the `X` feature matrix.

        The returned `X_block` is constructed by looping over each time in `feature_times`,
        finding the index position of that time in the series, and then for each lag value in
        `lags`, offset this index position by a particular lag value; this offset index is then
        used to extract all of the components at a single lagged time.

        Unlike the implementation found in `darts.utils.data.tabularization`, this function doesn't
        use any 'vectorisation' tricks, which makes it slower to run, but more easily interpretable.

        Some of the times in `feature_times` may occur before the start *or* after the end of `series`;
        see the docstrings of `get_feature_times_past` and `get_feature_times_future` for why this is the
        case. Because of this, we need to prepend or append these 'extended times' to `series.time_index`
        before searching for the index of each time in the series. Even though the integer indices of the
        'extended times' won't be contained within the original `series`, offsetting these found indices
        by the requested lag value should 'bring us back' to a time within the original, unextended `series`.
        However, if we've prepended times to `series.time_index`, we have to note that all of the indices will
        be 'bumped up' by the number of values we've prepended, even after offsetting by a lag value. For example,
        if we extended `series.time_index` by prepending two values to the start, the integer index of the first
        actual value in `series` will occur at an index of `2` instead of `0`. To 'undo' this, we must subtract off
        the number of prepended value from the lag-offseted indices before retrieving values from `series`.
        """
        if lags is None:
            X_block = None
        else:
            series_times = series.time_index
            is_range_idx = isinstance(series_times[0], int)
            # If series ends before last time in `feature_times`:
            add_to_end = series_times[-1] < feature_times[-1]
            # If series starts after first time in `feature_times`:
            add_to_start = series_times[0] > feature_times[0]
            # Either add to start *OR* add to end, we can't have both (see `get_feature_times_future`):
            if add_to_end:
                num_prepended = 0
                if is_range_idx:
                    # `+ 1` since `stop` index is exclusive:
                    series_times = pd.RangeIndex(
                        start=series_times[0],
                        stop=feature_times[-1] + 1,
                        step=series.freq,
                    )
                else:
                    series_times = pd.date_range(
                        start=series_times[0], end=feature_times[-1], freq=series.freq
                    )
            elif add_to_start:
                num_prepended = (series_times[0] - feature_times[0]) // series.freq
                if is_range_idx:
                    # `+ 1` since `stop` index is exclusive:
                    series_times = pd.RangeIndex(
                        start=feature_times[0],
                        stop=series_times[-1] + 1,
                        step=series.freq,
                    )
                else:
                    series_times = pd.date_range(
                        start=feature_times[0], end=series_times[-1], freq=series.freq
                    )
            else:
                num_prepended = 0
            array_vals = series.all_values(copy=False)[:, :, 0]
            X_block = []
            for time in feature_times:
                # Find position of time within series:
                time_idx = np.searchsorted(series_times, time)
                X_row = []
                for lag in lags:
                    # Offset by particular lag value:
                    idx_to_get = time_idx + lag
                    # Account for prepended values:
                    idx_to_get -= num_prepended
                    raise_if_not(
                        idx_to_get >= 0,
                        f"Unexpected case encountered: `time_idx + lag - num_prepended = {idx_to_get} < 0`.",
                    )
                    # Extract all components at this lagged time:
                    X_row.append(array_vals[idx_to_get, :].reshape(-1))
                # Concatenate together all lagged values into a single row:
                X_row = np.concatenate(X_row, axis=0)
                X_block.append(X_row)
            # Concatenate all rows (i.e. observations) together into a block:
            X_block = np.stack(X_block, axis=0)
        return X_block

    #
    #   Generated Test Cases
    #

    # Input parameter combinations used to generate test cases:
    max_samples_per_ts_combos = (1, 2, None)
    target_lag_combos = past_lag_combos = (None, [-1, -3], [-3, -1])
    future_lag_combos = (*target_lag_combos, [0], [2, 1], [-1, 1], [0, 2])

    @pytest.mark.parametrize(
        "series_type",
        ["datetime", "integer"],
    )
    def test_lagged_prediction_data_equal_freq(self, series_type):
        """
        Tests that `create_lagged_prediction_data` produces `X` and `times`
        outputs that are consistent with those generated by using the helper
        functions `get_feature_times` and `construct_X_block`. Consistency is
        checked over all of the combinations of parameter values specified by
        `self.target_lag_combos`, `self.covariates_lag_combos`, and
        `self.max_samples_per_ts_combos`.

        This particular test uses timeseries with time indices of equal
        frequencies. Since all of the timeseries are of the same frequency,
        the implementation of the 'moving window' method is being tested here.
        """
        # Define range index timeseries - each has different number of components,
        # different start times, different lengths, and different values, but
        # they're all of the same frequency:
        if series_type == "integer":
            target = self.create_multivariate_linear_timeseries(
                n_components=2, start_value=0, end_value=10, start=2, end=20, freq=2
            )
            past = self.create_multivariate_linear_timeseries(
                n_components=3, start_value=10, end_value=20, start=4, end=23, freq=2
            )
            future = self.create_multivariate_linear_timeseries(
                n_components=4, start_value=20, end_value=30, start=6, end=26, freq=2
            )
        else:
            target = self.create_multivariate_linear_timeseries(
                n_components=2,
                start_value=0,
                end_value=10,
                start=pd.Timestamp("1/2/2000"),
                end=pd.Timestamp("1/16/2000"),
                freq="2d",
            )
            past = self.create_multivariate_linear_timeseries(
                n_components=3,
                start_value=10,
                end_value=20,
                start=pd.Timestamp("1/4/2000"),
                end=pd.Timestamp("1/18/2000"),
                freq="2d",
            )
            future = self.create_multivariate_linear_timeseries(
                n_components=4,
                start_value=20,
                end_value=30,
                start=pd.Timestamp("1/6/2000"),
                end=pd.Timestamp("1/20/2000"),
                freq="2d",
            )
        # Conduct test for each input parameter combo:
        for lags, lags_past, lags_future, max_samples_per_ts in product(
            self.target_lag_combos,
            self.past_lag_combos,
            self.future_lag_combos,
            self.max_samples_per_ts_combos,
        ):
            all_lags = (lags, lags_past, lags_future)
            # Skip test where all lags are `None` - can't assemble features
            # for this single combo of input params:
            lags_is_none = [x is None for x in all_lags]
            if all(lags_is_none):
                continue
            X, times = create_lagged_prediction_data(
                target_series=target if lags else None,
                past_covariates=past if lags_past else None,
                future_covariates=future if lags_future else None,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                uses_static_covariates=False,
                max_samples_per_ts=max_samples_per_ts,
                use_moving_windows=True,
            )
            feats_times = self.get_feature_times(
                target,
                past,
                future,
                lags,
                lags_past,
                lags_future,
                max_samples_per_ts,
            )
            # Construct `X` by constructing each block, then concatenate these
            # blocks together along component axis:
            X_target = self.construct_X_block(target, feats_times, lags)
            X_past = self.construct_X_block(past, feats_times, lags_past)
            X_future = self.construct_X_block(future, feats_times, lags_future)
            all_X = (X_target, X_past, X_future)
            to_concat = [X for X in all_X if X is not None]
            expected_X = np.concatenate(to_concat, axis=1)
            # Number of observations should match number of feature times:
            assert X.shape[0] == len(feats_times)
            assert X.shape[0] == len(times[0])
            # Check that outputs match:
            assert np.allclose(expected_X, X[:, :, 0])
            assert feats_times.equals(times[0])

    @pytest.mark.parametrize(
        "series_type",
        ["datetime", "integer"],
    )
    def test_lagged_prediction_data_unequal_freq(self, series_type):
        """
        Tests that `create_lagged_prediction_data` produces `X` and `times`
        outputs that are consistent with those generated by using the helper
        functions `get_feature_times` and `construct_X_block`. Consistency is
        checked over all of the combinations of parameter values specified by
        `self.target_lag_combos`, `self.covariates_lag_combos`, and
        `self.max_samples_per_ts_combos`.

        This particular test uses timeseries with time indices of unequal
        frequencies. Since all of the timeseries are *not* of the same frequency,
        the implementation of the 'time intersection' method is being tested here.
        """
        # Define range index timeseries - each has different number of components,
        # different start times, different lengths, different values, and of
        # different frequencies:
        if series_type == "integer":
            target = self.create_multivariate_linear_timeseries(
                n_components=2, start_value=0, end_value=10, start=2, end=20, freq=1
            )
            past = self.create_multivariate_linear_timeseries(
                n_components=3, start_value=10, end_value=20, start=4, end=23, freq=2
            )
            future = self.create_multivariate_linear_timeseries(
                n_components=4, start_value=20, end_value=30, start=6, end=26, freq=3
            )
        else:
            target = self.create_multivariate_linear_timeseries(
                n_components=2,
                start_value=0,
                end_value=10,
                start=pd.Timestamp("1/2/2000"),
                end=pd.Timestamp("1/20/2000"),
                freq="1d",
            )
            past = self.create_multivariate_linear_timeseries(
                n_components=3,
                start_value=10,
                end_value=20,
                start=pd.Timestamp("1/4/2000"),
                end=pd.Timestamp("1/23/2000"),
                freq="2d",
            )
            future = self.create_multivariate_linear_timeseries(
                n_components=4,
                start_value=20,
                end_value=30,
                start=pd.Timestamp("1/6/2000"),
                end=pd.Timestamp("1/26/2000"),
                freq="3d",
            )
        # Conduct test for each input parameter combo:
        for lags, lags_past, lags_future, max_samples_per_ts in product(
            self.target_lag_combos,
            self.past_lag_combos,
            self.future_lag_combos,
            self.max_samples_per_ts_combos,
        ):
            all_lags = (lags, lags_past, lags_future)
            # Skip test where all lags are `None` - can't assemble features
            # for this single combo of input params:
            lags_is_none = [x is None for x in all_lags]
            if all(lags_is_none):
                continue
            X, times = create_lagged_prediction_data(
                target_series=target if lags else None,
                past_covariates=past if lags_past else None,
                future_covariates=future if lags_future else None,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                uses_static_covariates=False,
                max_samples_per_ts=max_samples_per_ts,
                use_moving_windows=True,
            )
            feats_times = self.get_feature_times(
                target,
                past,
                future,
                lags,
                lags_past,
                lags_future,
                max_samples_per_ts,
            )
            # Construct `X` by constructing each block, then concatenate these
            # blocks together along component axis:
            X_target = self.construct_X_block(target, feats_times, lags)
            X_past = self.construct_X_block(past, feats_times, lags_past)
            X_future = self.construct_X_block(future, feats_times, lags_future)
            all_X = (X_target, X_past, X_future)
            to_concat = [X for X in all_X if X is not None]
            expected_X = np.concatenate(to_concat, axis=1)
            # Number of observations should match number of feature times:
            assert X.shape[0] == len(feats_times)
            assert X.shape[0] == len(times[0])
            # Check that outputs match:
            assert np.allclose(expected_X, X[:, :, 0])
            assert feats_times.equals(times[0])

    @pytest.mark.parametrize(
        "series_type",
        ["datetime", "integer"],
    )
    def test_lagged_prediction_data_method_consistency_range_index(self, series_type):
        """
        Tests that `create_lagged_prediction_data` produces the same result
        when `use_moving_windows = False` and when `use_moving_windows = True`
        for all of the parameter combinations used in the 'generated' test cases.

        Obviously, if both the 'Moving Window Method' and the 'Time Intersection'
        are both wrong in the same way, this test won't reveal any bugs. With this
        being said, if this test fails, something is definitely wrong in either
        one or both of the implemented methods.
        """
        # Define datetime index timeseries - each has different number of components,
        # different start times, different lengths, different values, and of
        # different frequencies:
        if series_type == "integer":
            target = self.create_multivariate_linear_timeseries(
                n_components=2, start_value=0, end_value=10, start=2, end=16, freq=2
            )
            past = self.create_multivariate_linear_timeseries(
                n_components=3, start_value=10, end_value=20, start=4, end=18, freq=2
            )
            future = self.create_multivariate_linear_timeseries(
                n_components=4, start_value=20, end_value=30, start=6, end=20, freq=2
            )
        else:
            target = self.create_multivariate_linear_timeseries(
                n_components=2,
                start_value=0,
                end_value=10,
                start=pd.Timestamp("1/2/2000"),
                end=pd.Timestamp("1/16/2000"),
                freq="2d",
            )
            past = self.create_multivariate_linear_timeseries(
                n_components=3,
                start_value=10,
                end_value=20,
                start=pd.Timestamp("1/4/2000"),
                end=pd.Timestamp("1/18/2000"),
                freq="2d",
            )
            future = self.create_multivariate_linear_timeseries(
                n_components=4,
                start_value=20,
                end_value=30,
                start=pd.Timestamp("1/6/2000"),
                end=pd.Timestamp("1/20/2000"),
                freq="2d",
            )
        # Conduct test for each input parameter combo:
        for lags, lags_past, lags_future, max_samples_per_ts in product(
            self.target_lag_combos,
            self.past_lag_combos,
            self.future_lag_combos,
            self.max_samples_per_ts_combos,
        ):
            all_lags = (lags, lags_past, lags_future)
            # Skip test where all lags are `None` - can't assemble features
            # for this single combo of input params:
            lags_is_none = [x is None for x in all_lags]
            if all(lags_is_none):
                continue
            # Using moving window method:
            X_mw, times_mw = create_lagged_prediction_data(
                target_series=target if lags else None,
                past_covariates=past if lags_past else None,
                future_covariates=future if lags_future else None,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                uses_static_covariates=False,
                max_samples_per_ts=max_samples_per_ts,
                use_moving_windows=True,
            )
            # Using time intersection method:
            X_ti, times_ti = create_lagged_prediction_data(
                target_series=target if lags else None,
                past_covariates=past if lags_past else None,
                future_covariates=future if lags_future else None,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                uses_static_covariates=False,
                max_samples_per_ts=max_samples_per_ts,
                use_moving_windows=False,
            )
            assert np.allclose(X_mw, X_ti)
            assert times_mw[0].equals(times_ti[0])

    #
    #   Specified Cases Tests
    #

    @pytest.mark.parametrize(
        "config",
        itertools.product(["datetime", "integer"], [False, True]),
    )
    def test_lagged_prediction_data_single_lag_single_component_same_series(
        self, config
    ):
        """
        Tests that `create_lagged_prediction_data` correctly produces `X` and `times`
        when all the `series` inputs are identical, and all the `lags` inputs consist
        of a single value. In this situation, the expected `X` value can be found by
        concatenating three different slices of the same time series.
        """
        series_type, use_moving_windows = config
        if series_type == "integer":
            series = linear_timeseries(start=0, length=15)
        else:
            series = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=15)
        lags = [-1]
        past_lags = [-3]
        future_lags = [2]
        # Can't create features for first 3 times (because `past_lags`) and last
        # two times (because `future_lags`):
        expected_times = series.time_index[3:-2]
        # Offset `3:-2` by `-1` lag:
        expected_X_target = series.all_values(copy=False)[2:-3, :, 0]
        # Offset `3:-2` by `-3` lag -> gives `0:-5`:
        expected_X_past = series.all_values(copy=False)[:-5, :, 0]
        # Offset `3:-2` by `+2` lag -> gives `5:None`:
        expected_X_future = series.all_values(copy=False)[5:, :, 0]
        expected_X = np.concatenate(
            [expected_X_target, expected_X_past, expected_X_future], axis=1
        )
        X, times = create_lagged_prediction_data(
            target_series=series,
            past_covariates=series,
            future_covariates=series,
            lags=lags,
            lags_past_covariates=past_lags,
            lags_future_covariates=future_lags,
            uses_static_covariates=False,
            use_moving_windows=use_moving_windows,
        )
        # Number of observations should match number of feature times:
        assert X.shape[0] == len(expected_times)
        assert X.shape[0] == len(times[0])
        # Check that outputs match:
        assert np.allclose(expected_X, X[:, :, 0])
        assert expected_times.equals(times[0])

    @pytest.mark.parametrize(
        "config",
        itertools.product(["datetime", "integer"], [False, True]),
    )
    def test_lagged_prediction_data_extend_past_and_future_covariates(self, config):
        """
        Tests that `create_lagged_prediction_data` correctly handles case where features
        can be created for a time that is *not* contained in `target_series`, `past_covariates`
        and/or `future_covariates`.

        More specifically, we define the series and lags such that a prediction feature
        can be generated for time `target.end_time() + target.freq`, even though this time
        isn't contained in any of the define series.
        """
        series_type, use_moving_windows = config
        if series_type == "integer":
            # Can create feature for time `t = 9`, but this time isn't in any of the three series:
            target = linear_timeseries(start=0, end=9, start_value=1, end_value=2)
            past = linear_timeseries(start=0, end=8, start_value=2, end_value=3)
            future = linear_timeseries(start=0, end=6, start_value=3, end_value=4)
        else:
            # Can create feature for time `t = '1/10/2000'`, but this time isn't in any of the three series:
            target = linear_timeseries(
                start=pd.Timestamp("1/1/2000"),
                end=pd.Timestamp("1/10/2000"),
                start_value=1,
                end_value=2,
            )
            past = linear_timeseries(
                start=pd.Timestamp("1/1/2000"),
                end=pd.Timestamp("1/9/2000"),
                start_value=2,
                end_value=3,
            )
            future = linear_timeseries(
                start=pd.Timestamp("1/1/2000"),
                end=pd.Timestamp("1/7/2000"),
                start_value=3,
                end_value=4,
            )

        lags = [-1]
        lags_past = [-2]
        lags_future = [-4]
        # Only want to check very last generated observation:
        max_samples_per_ts = 1
        # Expect `X` to be constructed from the very last values of each series:
        expected_X = np.concatenate([
            target.all_values(copy=False)[-1, :, 0],
            past.all_values(copy=False)[-1, :, 0],
            future.all_values(copy=False)[-1, :, 0],
        ]).reshape(1, -1)
        # Check correctness for both 'moving window' method
        # and 'time intersection' method:
        X, times = create_lagged_prediction_data(
            target,
            past_covariates=past,
            future_covariates=future,
            lags=lags,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future,
            uses_static_covariates=False,
            max_samples_per_ts=max_samples_per_ts,
            use_moving_windows=use_moving_windows,
        )
        assert times[0][0] == target.end_time() + target.freq
        assert np.allclose(expected_X, X[:, :, 0])

    @pytest.mark.parametrize(
        "config",
        itertools.product(["datetime", "integer"], [False, True]),
    )
    def test_lagged_prediction_data_single_point(self, config):
        """
        Tests that `create_lagged_prediction_data` correctly handles case
        where only one possible training point can be generated.
        """
        # Can only create feature using first value of target (i.e. `0`):
        series_type, use_moving_windows = config
        if series_type == "integer":
            target = linear_timeseries(start=0, length=1, start_value=0, end_value=1)
        else:
            target = linear_timeseries(
                start=pd.Timestamp("1/1/2000"), length=1, start_value=0, end_value=1
            )

        expected_X = np.zeros((1, 1, 1))
        # Prediction time extend beyond end of series:
        lag = 5
        X, times = create_lagged_prediction_data(
            target,
            lags=[-lag],
            use_moving_windows=use_moving_windows,
            uses_static_covariates=False,
        )
        assert np.allclose(expected_X, X)
        # Should only have one sample, generated for
        # `t = target.end_time() + lag * target.freq`:
        assert len(times) == 1
        assert times[0] == target.end_time() + lag * target.freq

    @pytest.mark.parametrize(
        "config",
        itertools.product(["datetime", "integer"], [False, True]),
    )
    def test_lagged_prediction_data_zero_lags(self, config):
        """
        Tests that `create_lagged_prediction_data` correctly handles case when
        `0` is included in `lags_future_covariates` (i.e. when we're using the values
        `future_covariates` at time `t` to predict the value of `target_series` at
        that same time point).
        """
        # Define `future` so that only value occurs at the same time as
        # the only possible label that can be extracted from `target_series`; the
        # only possible feature that can be created using these series utilises
        # the value of `future` at the same time as the label (i.e. a lag
        # of `0` away from the only feature time):
        series_type, use_moving_windows = config
        if series_type == "integer":
            target = linear_timeseries(start=0, length=1, start_value=0, end_value=1)
            future = linear_timeseries(start=1, length=1, start_value=1, end_value=2)
        else:
            target = linear_timeseries(
                start=pd.Timestamp("1/1/2000"), length=1, start_value=0, end_value=1
            )
            future = linear_timeseries(
                start=pd.Timestamp("1/2/2000"), length=1, start_value=1, end_value=2
            )
        # X comprises of first value of `target` (i.e. 0) and only value in `future`:
        expected_X = np.array([0.0, 1.0]).reshape(1, 2, 1)
        # Check correctness for 'moving windows' and 'time intersection' methods, as
        # well as for different `multi_models` values:
        X, times = create_lagged_prediction_data(
            target,
            future_covariates=future,
            lags=[-1],
            lags_future_covariates=[0],
            uses_static_covariates=False,
            use_moving_windows=use_moving_windows,
        )
        assert np.allclose(expected_X, X)
        assert len(times[0]) == 1
        assert times[0][0] == future.start_time()

    @pytest.mark.parametrize(
        "config",
        itertools.product(["datetime", "integer"], [False, True]),
    )
    def test_lagged_prediction_data_positive_lags(self, config):
        """
        Tests that `create_lagged_prediction_data` correctly handles case when
        `0` is included in `lags_future_covariates` (i.e. when we're using the values
        `future_covariates` at time `t` to predict the value of `target_series` at
        that same time point). This particular test checks this behaviour by using
        datetime index timeseries.
        """
        # Define `future` so that only value occurs one timestep *after*
        # the only possible label that can be extracted from `target_series`; the
        # only possible feature that can be created using these series utilises
        # the value of `future` one timestep after the time of the label (i.e. a lag
        # of `1` away from the only feature time):
        series_type, use_moving_windows = config
        if series_type == "integer":
            target = linear_timeseries(start=0, length=1, start_value=0, end_value=1)
            future = linear_timeseries(start=2, length=1, start_value=1, end_value=2)
        else:
            target = linear_timeseries(
                start=pd.Timestamp("1/1/2000"), length=1, start_value=0, end_value=1
            )
            future = linear_timeseries(
                start=pd.Timestamp("1/3/2000"), length=1, start_value=1, end_value=2
            )
        # X comprises of first value of `target` (i.e. 0) and only value in `future`:
        expected_X = np.array([0.0, 1.0]).reshape(1, 2, 1)
        # Check correctness for 'moving windows' and 'time intersection' methods, as
        # well as for different `multi_models` values:
        X, times = create_lagged_prediction_data(
            target,
            future_covariates=future,
            lags=[-1],
            lags_future_covariates=[1],
            uses_static_covariates=False,
            use_moving_windows=use_moving_windows,
        )
        assert np.allclose(expected_X, X)
        assert len(times[0]) == 1
        assert times[0][0] == target.end_time() + target.freq

    def test_lagged_prediction_data_sequence_inputs(self):
        """
        Tests that `create_lagged_prediction_data` correctly handles being
        passed a sequence of `TimeSeries` inputs, as opposed to individual
        `TimeSeries`.
        """
        # Define two simple tabularization problems:
        target_1 = past_1 = future_1 = linear_timeseries(start=0, end=5)
        target_2 = past_2 = future_2 = linear_timeseries(start=6, end=11)
        lags = lags_past = lags_future = [-1]
        # Expected solution:
        expected_X_1 = np.concatenate(3 * [target_1.all_values(copy=False)], axis=1)
        expected_X_2 = np.concatenate(3 * [target_2.all_values(copy=False)], axis=1)
        expected_X = np.concatenate([expected_X_1, expected_X_2], axis=0)
        # Need to account for fact that we can create feature for last time in series:
        expected_times_1 = target_1.append_values([0]).time_index[1:]
        expected_times_2 = target_2.append_values([0]).time_index[1:]
        # Check when `concatenate = True`:
        X, times = create_lagged_prediction_data(
            target_series=(target_1, target_2),
            past_covariates=(past_1, past_2),
            future_covariates=(future_1, future_2),
            lags=lags,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future,
            uses_static_covariates=False,
        )
        assert np.allclose(X, expected_X)
        assert len(times) == 2
        assert times[0].equals(expected_times_1)
        assert times[1].equals(expected_times_2)
        # Check when `concatenate = False`:
        X, times = create_lagged_prediction_data(
            target_series=(target_1, target_2),
            past_covariates=(past_1, past_2),
            future_covariates=(future_1, future_2),
            lags=lags,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future,
            uses_static_covariates=False,
            concatenate=False,
        )
        assert len(X) == 2
        assert np.allclose(X[0], expected_X_1)
        assert np.allclose(X[1], expected_X_2)
        assert len(times) == 2
        assert times[0].equals(expected_times_1)
        assert times[1].equals(expected_times_2)

    def test_lagged_prediction_data_stochastic_series(self):
        """
        Tests that `create_lagged_prediction_data` is correctly vectorised
        over the sample axes of the input `TimeSeries`.
        """
        # Define two simple tabularization problems:
        target_1 = past_1 = future_1 = linear_timeseries(start=0, end=5)
        target_2 = past_2 = future_2 = 2 * target_1
        target = target_1.concatenate(target_2, axis=2)
        past = past_1.concatenate(past_2, axis=2)
        future = future_1.concatenate(future_2, axis=2)
        lags = lags_past = lags_future = [-1]
        # Expected solution:
        expected_X = np.concatenate(3 * [target.all_values(copy=False)], axis=1)
        # Need to account for fact that we can create feature for last time in series:
        expected_times = target_1.append_values([0]).time_index[1:]
        X, times = create_lagged_prediction_data(
            target_series=target,
            past_covariates=past,
            future_covariates=future,
            lags=lags,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future,
            uses_static_covariates=False,
        )
        assert np.allclose(X, expected_X)
        assert times[0].equals(expected_times)

    def test_lagged_prediction_data_no_shared_times_error(self):
        """
        Tests that `create_lagged_prediction_data` throws correct error
        when the specified series do not share any times in common
        for creating features and labels.
        """
        # Have `series_1` share no overlap with `series_2`:
        series_1 = linear_timeseries(start=0, length=4, freq=1)
        series_2 = linear_timeseries(start=series_1.end_time() + 1, length=4, freq=1)
        lags = [-1]
        # Check error thrown by 'moving windows' method and by 'time intersection' method:
        for use_moving_windows in (False, True):
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    target_series=series_1,
                    lags=lags,
                    past_covariates=series_2,
                    lags_past_covariates=lags,
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
            assert (
                "Specified series do not share any common times for which features can be created."
                == str(err.value)
            )
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    target_series=series_1,
                    lags=lags,
                    past_covariates=series_2,
                    lags_past_covariates=lags,
                    uses_static_covariates=False,
                )
            assert (
                "Specified series do not share any common times for which features can be created."
                == str(err.value)
            )

    def test_lagged_prediction_data_no_specified_series_lags_pairs_error(self):
        """
        Tests that `create_lagged_prediction_data` throws correct error
        when no lags-series pairs are specified.
        """
        # Define some arbitrary inputs:
        series = linear_timeseries(start=1, length=10, freq=1)
        # Check error thrown by 'moving windows' method and by 'time intersection' method:
        for use_moving_windows in (False, True):
            # Three warnings will be thrown indicating that series
            # is specified without corresponding lags or vice
            # versa - ignore these warnings:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                with pytest.raises(ValueError) as err:
                    create_lagged_prediction_data(
                        target_series=series,
                        lags_future_covariates=[-1],
                        past_covariates=series,
                        uses_static_covariates=False,
                        use_moving_windows=use_moving_windows,
                    )
            assert "Must specify at least one series-lags pair." == str(err.value)

    def test_lagged_prediction_data_no_lags_specified_error(self):
        """
        Tests that `create_lagged_prediction_data` throws correct error
        when no lags are specified.
        """
        target = linear_timeseries(start=1, length=20, freq=1)
        # Check error thrown by 'moving windows' method and by 'time intersection' method:
        for use_moving_windows in (False, True):
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    target_series=target,
                    use_moving_windows=use_moving_windows,
                    uses_static_covariates=False,
                )
            assert (
                "Must specify at least one of: `lags`, `lags_past_covariates`, `lags_future_covariates`."
                == str(err.value)
            )

    def test_lagged_prediction_data_series_too_short_error(self):
        """
        Tests that `create_lagged_prediction_data` throws correct error
        when supplied series is too short to generate any
        features from using the specified `lags`.
        """
        series = linear_timeseries(start=1, length=2, freq=1)
        # Check error thrown by 'moving windows' method and by 'time intersection' method:
        for use_moving_windows in (False, True):
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    target_series=series,
                    lags=[-20, -1],
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
            assert (
                "`target_series` must have at least "
                "`-min(lags) + max(lags) + 1` = 20 "
                "time steps; instead, it only has 2."
            ) == str(err.value)
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    past_covariates=series,
                    lags_past_covariates=[-20, -1],
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
            assert (
                "`past_covariates` must have at least "
                "`-min(lags_past_covariates) + max(lags_past_covariates) + 1` = 20 "
                "time steps; instead, it only has 2."
            ) == str(err.value)

    def test_lagged_prediction_data_invalid_lag_values_error(self):
        """
        Tests that `create_lagged_prediction_data` throws correct
        error when invalid lag values are specified. More specifically:
            - If `lags` contains any value greater than `-1`, an error
            should be thrown (since times > `-1` are used for labels).
            - If `lags_past_covariates`/`lags_future_covariates` contains
            any value greater than `0`, an error should be thrown (i.e. we
            shouldn't be predicting the past using the future).
        """
        series = linear_timeseries(start=1, length=3, freq=1)
        # Check error thrown by 'moving windows' method and by 'time intersection' method:
        for use_moving_windows in (False, True):
            # Test invalid `lags` values:
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    target_series=series,
                    lags=[0],
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
            assert (
                "`lags` must be a `Sequence` or `Dict` containing only `int` values less than 0."
            ) == str(err.value)
            # Test invalid `lags_past_covariates` values:
            with pytest.raises(ValueError) as err:
                create_lagged_prediction_data(
                    past_covariates=series,
                    lags_past_covariates=[0],
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
            assert (
                "`lags_past_covariates` must be a `Sequence` or `Dict` containing only `int` values less than 0."
            ) == str(err.value)
            # This should *not* throw an error:
            create_lagged_prediction_data(
                future_covariates=series,
                lags_future_covariates=[-1, 0, 1],
                uses_static_covariates=False,
                use_moving_windows=use_moving_windows,
            )

    def test_lagged_prediction_data_unspecified_lag_or_series_warning(self):
        """
        Tests that `create_lagged_prediction_data` throws correct
        user warnings when a series is specified without any
        corresponding lags, or vice versa.
        """
        series = linear_timeseries(start=1, length=20, freq=1)
        lags = [-1]
        # Check warnings thrown by 'moving windows' method and by 'time intersection' method:
        for use_moving_windows in (False, True):
            with warnings.catch_warnings(record=True) as w:
                _ = create_lagged_prediction_data(
                    target_series=series,
                    lags=lags,
                    future_covariates=series,
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
                assert len(w) == 1
                assert issubclass(w[0].category, UserWarning)
                assert str(w[0].message) == (
                    "`future_covariates` was specified without accompanying "
                    "`lags_future_covariates` and, thus, will be ignored."
                )
            # Specify `lags_future_covariates` but not `future_covariates`:
            with warnings.catch_warnings(record=True) as w:
                _ = create_lagged_prediction_data(
                    target_series=series,
                    lags=lags,
                    lags_future_covariates=lags,
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
                assert len(w) == 1
                assert issubclass(w[0].category, UserWarning)
                assert str(w[0].message) == (
                    "`lags_future_covariates` was specified without accompanying "
                    "`future_covariates` and, thus, will be ignored."
                )
            # Specify `lags_future_covariates` but not `future_covariates`, and
            # `past_covariates` but not `lags_past_covariates`, and `target_series`
            # but not `lags`:
            with warnings.catch_warnings(record=True) as w:
                _ = create_lagged_prediction_data(
                    target_series=series,
                    lags=lags,
                    past_covariates=series,
                    lags_future_covariates=lags,
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
                assert len(w) == 2
                assert issubclass(w[0].category, UserWarning)
                assert issubclass(w[1].category, UserWarning)
                assert str(w[0].message) == (
                    "`past_covariates` was specified without accompanying "
                    "`lags_past_covariates` and, thus, will be ignored."
                )
                assert str(w[1].message) == (
                    "`lags_future_covariates` was specified without accompanying "
                    "`future_covariates` and, thus, will be ignored."
                )
            # Specify `target_series` but not `lags` - this *should* throw
            # a warning when creating prediction data:
            with warnings.catch_warnings(record=True) as w:
                _ = create_lagged_prediction_data(
                    target_series=series,
                    past_covariates=series,
                    lags_past_covariates=lags,
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
                assert len(w) == 1
                assert issubclass(w[0].category, UserWarning)
                assert str(w[0].message) == (
                    "`target_series` was specified without accompanying "
                    "`lags` and, thus, will be ignored."
                )
            # Specify `target_series` but not `lags` - this *should* throw
            # a warning when creating prediction data:
            with warnings.catch_warnings(record=True) as w:
                _ = create_lagged_prediction_data(
                    lags=lags,
                    past_covariates=series,
                    lags_past_covariates=lags,
                    uses_static_covariates=False,
                    use_moving_windows=use_moving_windows,
                )
                assert len(w) == 1
                assert issubclass(w[0].category, UserWarning)
                assert str(w[0].message) == (
                    "`lags` was specified without accompanying "
                    "`target_series` and, thus, will be ignored."
                )



import copy
import itertools
import os
from typing import Any, Optional
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

import darts.utils.timeseries_generation as tg
from darts import TimeSeries
from darts.dataprocessing.encoders import SequentialEncoder
from darts.dataprocessing.transformers import BoxCox, Scaler
from darts.metrics import mape
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )
import torch
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.loggers.logger import DummyLogger
from pytorch_lightning.tuner.lr_finder import _LRFinder
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from torchmetrics import (
    MeanAbsoluteError,
    MeanAbsolutePercentageError,
    Metric,
    MetricCollection,
)

from darts.models import (
    BlockRNNModel,
    DLinearModel,
    GlobalNaiveAggregate,
    GlobalNaiveDrift,
    GlobalNaiveSeasonal,
    NBEATSModel,
    NHiTSModel,
    NLinearModel,
    RNNModel,
    TCNModel,
    TFTModel,
    TiDEModel,
    TransformerModel,
    TSMixerModel,
)
from darts.models.components.layer_norm_variants import RINorm
from darts.models.forecasting.global_baseline_models import _GlobalNaiveModel
from darts.utils.likelihood_models import (
    CauchyLikelihood,
    GaussianLikelihood,
    LaplaceLikelihood,
    Likelihood,
    QuantileRegression,
)

kwargs = {
    "input_chunk_length": 10,
    "output_chunk_length": 1,
    "n_epochs": 1,
    "random_state": 42,
    "pl_trainer_kwargs": {"fast_dev_run": True, **tfm_kwargs["pl_trainer_kwargs"]},
}
# make models light weight
dlinear_light_kwargs = {"kernel_size": 2}
nbeats_light_kwargs = {
    "num_stacks": 1,
    "num_blocks": 1,
    "num_layers": 1,
    "layer_widths": 2,
}
tcn_light_kwargs = {
    "kernel_size": 2,
    "num_filters": 1,
    "dilation_base": 1,
}
trafo_light_kwargs = {
    "d_model": 2,
    "nhead": 1,
    "num_encoder_layers": 1,
    "num_decoder_layers": 1,
    "dim_feedforward": 2,
}
tft_light_kwargs = {
    "hidden_size": 2,
    "lstm_layers": 1,
    "num_attention_heads": 1,
    "hidden_continuous_size": 2,
}
models = [
    (BlockRNNModel, kwargs),
    (DLinearModel, dict(kwargs, **dlinear_light_kwargs)),
    (NBEATSModel, dict(kwargs, **nbeats_light_kwargs)),
    (NHiTSModel, dict(kwargs, **nbeats_light_kwargs)),
    (NLinearModel, kwargs),
    (RNNModel, {"training_length": 10, **kwargs}),
    (TCNModel, dict(kwargs, **tcn_light_kwargs)),
    (TFTModel, {"add_relative_index": 2, **kwargs, **tft_light_kwargs}),
    (TiDEModel, kwargs),
    (TransformerModel, dict(kwargs, **trafo_light_kwargs)),
    (TSMixerModel, kwargs),
    (GlobalNaiveSeasonal, kwargs),
    (GlobalNaiveAggregate, kwargs),
    (GlobalNaiveDrift, kwargs),
]


class NumsCalled(Metric):
    def __init__(self):
        super().__init__()
        self.add_state("preds", default=[], dist_reduce_fx="cat")

    def update(self, preds, target) -> None:
        self.preds.append(preds)

    def compute(self):
        return len(self.preds)


class TestTorchForecastingModel:
    times = pd.date_range("20130101", "20130410")
    pd_series = pd.Series(range(100), index=times)
    series = TimeSeries.from_series(pd_series)

    df = pd.DataFrame({"var1": range(100), "var2": range(100)}, index=times)
    multivariate_series = TimeSeries.from_dataframe(df)

    def test_save_model_parameters(self):
        # check if re-created model has same params as original
        model = RNNModel(12, "RNN", 10, 10, **tfm_kwargs)
        params_old = model.model_params
        params_new = model.untrained_model().model_params

        assert params_old.keys() == params_new.keys()
        assert all([params_old[k] == params_new[k] for k in params_old])

    @patch(
        "darts.models.forecasting.torch_forecasting_model.TorchForecastingModel.save"
    )
    def test_suppress_automatic_save(self, patch_save_model, tmpdir_fn):
        model_name = "test_model"
        model1 = RNNModel(
            12,
            "RNN",
            10,
            10,
            model_name=model_name,
            work_dir=tmpdir_fn,
            save_checkpoints=False,
            **tfm_kwargs,
        )
        model2 = RNNModel(
            12,
            "RNN",
            10,
            10,
            model_name=model_name,
            work_dir=tmpdir_fn,
            force_reset=True,
            save_checkpoints=False,
            **tfm_kwargs,
        )

        model1.fit(self.series, epochs=1)
        model2.fit(self.series, epochs=1)

        model1.predict(n=1)
        model2.predict(n=2)

        patch_save_model.assert_not_called()

        model1.save(path=os.path.join(tmpdir_fn, model_name))
        patch_save_model.assert_called()

    def test_manual_save_and_load(self, tmpdir_fn):
        """validate manual save with automatic save files by comparing output between the two"""

        model_dir = os.path.join(tmpdir_fn)
        manual_name = "test_save_manual"
        auto_name = "test_save_automatic"
        model_manual_save = RNNModel(
            12,
            "RNN",
            10,
            10,
            model_name=manual_name,
            work_dir=tmpdir_fn,
            save_checkpoints=False,
            random_state=42,
            **tfm_kwargs,
        )
        model_auto_save = RNNModel(
            12,
            "RNN",
            10,
            10,
            model_name=auto_name,
            work_dir=tmpdir_fn,
            save_checkpoints=True,
            random_state=42,
            **tfm_kwargs,
        )

        # save model without training
        no_training_ckpt = "no_training.pth.tar"
        no_training_ckpt_path = os.path.join(model_dir, no_training_ckpt)
        model_manual_save.save(no_training_ckpt_path)
        # check that model object file was created
        assert os.path.exists(no_training_ckpt_path)
        # check that the PyTorch Ligthning ckpt does not exist
        assert not os.path.exists(no_training_ckpt_path + ".ckpt")
        # informative exception about `fit()` not called
        with pytest.raises(ValueError) as err:
            no_train_model = RNNModel.load(no_training_ckpt_path)
            no_train_model.predict(n=4)
        assert str(err.value) == (
            "Input `series` must be provided. This is the result either from "
            "fitting on multiple series, or from not having fit the model yet."
        )

        model_manual_save.fit(self.series, epochs=1)
        model_auto_save.fit(self.series, epochs=1)

        # check that file was not created with manual save
        assert not os.path.exists(os.path.join(model_dir, manual_name, "checkpoints"))
        # check that file was created with automatic save
        assert os.path.exists(os.path.join(model_dir, auto_name, "checkpoints"))

        # create manually saved model checkpoints folder
        checkpoint_path_manual = os.path.join(model_dir, manual_name)
        os.mkdir(checkpoint_path_manual)

        checkpoint_file_name = "checkpoint_0.pth.tar"
        model_path_manual = os.path.join(checkpoint_path_manual, checkpoint_file_name)
        checkpoint_file_name_cpkt = "checkpoint_0.pth.tar.ckpt"
        model_path_manual_ckpt = os.path.join(
            checkpoint_path_manual, checkpoint_file_name_cpkt
        )

        # save manually saved model
        model_manual_save.save(model_path_manual)
        assert os.path.exists(model_path_manual)

        # check that the PTL checkpoint path is also there
        assert os.path.exists(model_path_manual_ckpt)

        # load manual save model and compare with automatic model results
        model_manual_save = RNNModel.load(model_path_manual, map_location="cpu")
        model_manual_save.to_cpu()
        assert model_manual_save.predict(n=4) == model_auto_save.predict(n=4)

        # load automatically saved model with manual load() and load_from_checkpoint()
        model_auto_save1 = RNNModel.load_from_checkpoint(
            model_name=auto_name,
            work_dir=tmpdir_fn,
            best=False,
            map_location="cpu",
        )
        model_auto_save1.to_cpu()
        # compare loaded checkpoint with manual save
        assert model_manual_save.predict(n=4) == model_auto_save1.predict(n=4)

        # save() model directly after load_from_checkpoint()
        checkpoint_file_name_2 = "checkpoint_1.pth.tar"
        checkpoint_file_name_cpkt_2 = checkpoint_file_name_2 + ".ckpt"

        model_path_manual_2 = os.path.join(
            checkpoint_path_manual, checkpoint_file_name_2
        )
        model_path_manual_ckpt_2 = os.path.join(
            checkpoint_path_manual, checkpoint_file_name_cpkt_2
        )
        model_auto_save2 = RNNModel.load_from_checkpoint(
            model_name=auto_name,
            work_dir=tmpdir_fn,
            best=False,
            map_location="cpu",
        )
        # save model directly after loading, model has no trainer
        model_auto_save2.save(model_path_manual_2)

        # assert original .ckpt checkpoint was correctly copied
        assert os.path.exists(model_path_manual_ckpt_2)

        model_chained_load_save = RNNModel.load(model_path_manual_2, map_location="cpu")

        # compare chained load_from_checkpoint() save() with manual save
        assert model_chained_load_save.predict(n=4) == model_manual_save.predict(n=4)

    def test_valid_save_and_load_weights_with_different_params(self, tmpdir_fn):
        """
        Verify that save/load does not break encoders.

        Note: since load_weights() calls load_weights_from_checkpoint(), it will be used
        for all but one test.
        Note: Using DLinear since it supports both past and future covariates
        """

        def create_model(**kwargs):
            return DLinearModel(
                input_chunk_length=4,
                output_chunk_length=1,
                **kwargs,
                **tfm_kwargs,
            )

        model_dir = os.path.join(tmpdir_fn)
        manual_name = "save_manual"
        # create manually saved model checkpoints folder
        checkpoint_path_manual = os.path.join(model_dir, manual_name)
        os.mkdir(checkpoint_path_manual)
        checkpoint_file_name = "checkpoint_0.pth.tar"
        model_path_manual = os.path.join(checkpoint_path_manual, checkpoint_file_name)
        model = create_model()
        model.fit(self.series, epochs=1)
        model.save(model_path_manual)

        kwargs_valid = [
            {"optimizer_cls": torch.optim.SGD},
            {"optimizer_kwargs": {"lr": 0.1}},
        ]
        # check that all models can be created with different valid kwargs
        for kwargs_ in kwargs_valid:
            model_new = create_model(**kwargs_)
            model_new.load_weights(model_path_manual)

    def test_save_and_load_weights_w_encoders(self, tmpdir_fn):
        """
        Verify that save/load does not break encoders.

        Note: since load_weights() calls load_weights_from_checkpoint(), it will be used
        for all but one test.
        Note: Using DLinear since it supports both past and future covariates
        """
        model_dir = os.path.join(tmpdir_fn)
        manual_name = "save_manual"
        auto_name = "save_auto"
        auto_name_other = "save_auto_other"
        # create manually saved model checkpoints folder
        checkpoint_path_manual = os.path.join(model_dir, manual_name)
        os.mkdir(checkpoint_path_manual)
        checkpoint_file_name = "checkpoint_0.pth.tar"
        model_path_manual = os.path.join(checkpoint_path_manual, checkpoint_file_name)

        # define encoders sets
        encoders_past = {
            "datetime_attribute": {"past": ["day"]},
            "transformer": Scaler(),
        }
        encoders_other_past = {
            "datetime_attribute": {"past": ["hour"]},
            "transformer": Scaler(),
        }
        encoders_past_noscaler = {
            "datetime_attribute": {"past": ["day"]},
        }
        encoders_past_other_transformer = {
            "datetime_attribute": {"past": ["day"]},
            "transformer": BoxCox(lmbda=-0.7),
        }
        encoders_2_past = {
            "datetime_attribute": {"past": ["hour", "day"]},
            "transformer": Scaler(),
        }
        encoders_past_n_future = {
            "datetime_attribute": {"past": ["day"], "future": ["dayofweek"]},
            "transformer": Scaler(),
        }

        model_auto_save = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name=auto_name,
            save_checkpoints=True,
            add_encoders=encoders_past,
        )
        model_auto_save.fit(self.series, epochs=1)

        model_manual_save = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name=manual_name,
            save_checkpoints=False,
            add_encoders=encoders_past,
        )
        model_manual_save.fit(self.series, epochs=1)
        model_manual_save.save(model_path_manual)

        model_auto_save_other = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name=auto_name_other,
            save_checkpoints=True,
            add_encoders=encoders_other_past,
        )
        model_auto_save_other.fit(self.series, epochs=1)

        # prediction are different when using different encoders
        assert model_auto_save.predict(n=4) != model_auto_save_other.predict(n=4)

        # model with undeclared encoders
        model_no_enc = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn, model_name="no_encoder", add_encoders=None
        )
        # weights were trained with encoders, new model must be instantiated with encoders
        with pytest.raises(ValueError):
            model_no_enc.load_weights_from_checkpoint(
                auto_name,
                work_dir=tmpdir_fn,
                best=False,
                load_encoders=False,
                map_location="cpu",
            )
        # overwrite undeclared encoders
        model_no_enc.load_weights_from_checkpoint(
            auto_name,
            work_dir=tmpdir_fn,
            best=False,
            load_encoders=True,
            map_location="cpu",
        )
        self.helper_equality_encoders(
            model_auto_save.add_encoders, model_no_enc.add_encoders
        )
        self.helper_equality_encoders_transfo(
            model_auto_save.add_encoders, model_no_enc.add_encoders
        )
        # cannot directly verify equality between encoders, using predict as proxy
        assert model_auto_save.predict(n=4) == model_no_enc.predict(
            n=4, series=self.series
        )

        # model with identical encoders (fittable)
        model_same_enc_noload = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_encoder_noload",
            add_encoders=encoders_past,
        )
        model_same_enc_noload.load_weights(
            model_path_manual,
            load_encoders=False,
            map_location="cpu",
        )
        # cannot predict because of un-fitted encoder
        with pytest.raises(ValueError):
            model_same_enc_noload.predict(n=4, series=self.series)

        model_same_enc_load = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_encoder_load",
            add_encoders=encoders_past,
        )
        model_same_enc_load.load_weights(
            model_path_manual,
            load_encoders=True,
            map_location="cpu",
        )
        assert model_manual_save.predict(n=4) == model_same_enc_load.predict(
            n=4, series=self.series
        )

        # model with different encoders (fittable)
        model_other_enc_load = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="other_encoder_load",
            add_encoders=encoders_other_past,
        )
        # cannot overwrite different declared encoders
        with pytest.raises(ValueError):
            model_other_enc_load.load_weights(
                model_path_manual,
                load_encoders=True,
                map_location="cpu",
            )

        # model with different encoders but same dimensions (fittable)
        model_other_enc_noload = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="other_encoder_noload",
            add_encoders=encoders_other_past,
        )
        model_other_enc_noload.load_weights(
            model_path_manual,
            load_encoders=False,
            map_location="cpu",
        )
        self.helper_equality_encoders(
            model_other_enc_noload.add_encoders, encoders_other_past
        )
        self.helper_equality_encoders_transfo(
            model_other_enc_noload.add_encoders, encoders_other_past
        )
        # new encoders were instantiated
        assert isinstance(model_other_enc_noload.encoders, SequentialEncoder)
        # since fit() was not called, new fittable encoders were not trained
        with pytest.raises(ValueError):
            model_other_enc_noload.predict(n=4, series=self.series)

        # predict() can be called after fit()
        model_other_enc_noload.fit(self.series, epochs=1)
        model_other_enc_noload.predict(n=4, series=self.series)

        # model with same encoders but no scaler (non-fittable)
        model_new_enc_noscaler_noload = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_encoder_noscaler",
            add_encoders=encoders_past_noscaler,
        )
        model_new_enc_noscaler_noload.load_weights(
            model_path_manual,
            load_encoders=False,
            map_location="cpu",
        )

        self.helper_equality_encoders(
            model_new_enc_noscaler_noload.add_encoders, encoders_past_noscaler
        )
        self.helper_equality_encoders_transfo(
            model_new_enc_noscaler_noload.add_encoders, encoders_past_noscaler
        )
        # predict() can be called directly since new encoders don't contain scaler
        model_new_enc_noscaler_noload.predict(n=4, series=self.series)

        # model with same encoders but different transformer (fittable)
        model_new_enc_other_transformer = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_encoder_other_transform",
            add_encoders=encoders_past_other_transformer,
        )
        # cannot overwrite different declared encoders
        with pytest.raises(ValueError):
            model_new_enc_other_transformer.load_weights(
                model_path_manual,
                load_encoders=True,
                map_location="cpu",
            )

        model_new_enc_other_transformer.load_weights(
            model_path_manual,
            load_encoders=False,
            map_location="cpu",
        )
        # since fit() was not called, new fittable encoders were not trained
        with pytest.raises(ValueError):
            model_new_enc_other_transformer.predict(n=4, series=self.series)

        # predict() can be called after fit()
        model_new_enc_other_transformer.fit(self.series, epochs=1)
        model_new_enc_other_transformer.predict(n=4, series=self.series)

        # model with encoders containing more components (fittable)
        model_new_enc_2_past = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="encoder_2_components_past",
            add_encoders=encoders_2_past,
        )
        # cannot overwrite different declared encoders
        with pytest.raises(ValueError):
            model_new_enc_2_past.load_weights(
                model_path_manual,
                load_encoders=True,
                map_location="cpu",
            )
        # new encoders have one additional past component
        with pytest.raises(ValueError):
            model_new_enc_2_past.load_weights(
                model_path_manual,
                load_encoders=False,
                map_location="cpu",
            )

        # model with encoders containing past and future covs (fittable)
        model_new_enc_past_n_future = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="encoder_past_n_future",
            add_encoders=encoders_past_n_future,
        )
        # cannot overwrite different declared encoders
        with pytest.raises(ValueError):
            model_new_enc_past_n_future.load_weights(
                model_path_manual,
                load_encoders=True,
                map_location="cpu",
            )
        # identical past components, but different future components
        with pytest.raises(ValueError):
            model_new_enc_past_n_future.load_weights(
                model_path_manual,
                load_encoders=False,
                map_location="cpu",
            )

    def test_save_and_load_weights_w_likelihood(self, tmpdir_fn):
        """
        Verify that save/load does not break likelihood.

        Note: since load_weights() calls load_weights_from_checkpoint(), it will be used
        for all but one test.
        Note: Using DLinear since it supports both past and future covariates
        """
        model_dir = os.path.join(tmpdir_fn)
        manual_name = "save_manual"
        auto_name = "save_auto"
        # create manually saved model checkpoints folder
        checkpoint_path_manual = os.path.join(model_dir, manual_name)
        os.mkdir(checkpoint_path_manual)
        checkpoint_file_name = "checkpoint_0.pth.tar"
        model_path_manual = os.path.join(checkpoint_path_manual, checkpoint_file_name)

        model_auto_save = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name=auto_name,
            save_checkpoints=True,
            likelihood=GaussianLikelihood(prior_mu=0.5),
        )
        model_auto_save.fit(self.series, epochs=1)
        pred_auto = model_auto_save.predict(n=4, series=self.series)

        model_manual_save = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name=manual_name,
            save_checkpoints=False,
            likelihood=GaussianLikelihood(prior_mu=0.5),
        )
        model_manual_save.fit(self.series, epochs=1)
        model_manual_save.save(model_path_manual)
        pred_manual = model_manual_save.predict(n=4, series=self.series)

        # predictions are identical when using the same likelihood
        assert np.array_equal(pred_auto.values(), pred_manual.values())

        # model with identical likelihood
        model_same_likelihood = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_likelihood",
            likelihood=GaussianLikelihood(prior_mu=0.5),
        )
        model_same_likelihood.load_weights(model_path_manual, map_location="cpu")
        model_same_likelihood.predict(n=4, series=self.series)
        # cannot check predictions since this model is not fitted, random state is different

        # loading models weights with respective methods
        model_manual_same_likelihood = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_likelihood",
            likelihood=GaussianLikelihood(prior_mu=0.5),
        )
        model_manual_same_likelihood.load_weights(model_path_manual, map_location="cpu")
        preds_manual_from_weights = model_manual_same_likelihood.predict(
            n=4, series=self.series
        )

        model_auto_same_likelihood = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_likelihood",
            likelihood=GaussianLikelihood(prior_mu=0.5),
        )
        model_auto_same_likelihood.load_weights_from_checkpoint(
            auto_name, work_dir=tmpdir_fn, best=False, map_location="cpu"
        )
        preds_auto_from_weights = model_auto_same_likelihood.predict(
            n=4, series=self.series
        )
        # check that weights from checkpoint give identical predictions as weights from manual save
        assert preds_manual_from_weights == preds_auto_from_weights
        # model with explicitly no likelihood
        model_no_likelihood = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn, model_name="no_likelihood", likelihood=None
        )
        with pytest.raises(ValueError) as error_msg:
            model_no_likelihood.load_weights_from_checkpoint(
                auto_name,
                work_dir=tmpdir_fn,
                best=False,
                map_location="cpu",
            )
        assert str(error_msg.value).startswith(
            "The values of the hyper-parameters in the model and loaded checkpoint should be identical.\n"
            "incorrect"
        )

        # model with missing likelihood (as if user forgot them)
        model_no_likelihood_bis = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=1,
            model_name="no_likelihood_bis",
            add_encoders=None,
            work_dir=tmpdir_fn,
            save_checkpoints=False,
            random_state=42,
            force_reset=True,
            n_epochs=1,
            # likelihood=likelihood,
            **tfm_kwargs,
        )
        with pytest.raises(ValueError) as error_msg:
            model_no_likelihood_bis.load_weights_from_checkpoint(
                auto_name,
                work_dir=tmpdir_fn,
                best=False,
                map_location="cpu",
            )
        assert str(error_msg.value).startswith(
            "The values of the hyper-parameters in the model and loaded checkpoint should be identical.\n"
            "missing"
        )

        # model with a different likelihood
        model_other_likelihood = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="other_likelihood",
            likelihood=LaplaceLikelihood(),
        )
        with pytest.raises(ValueError) as error_msg:
            model_other_likelihood.load_weights(model_path_manual, map_location="cpu")
        assert str(error_msg.value).startswith(
            "The values of the hyper-parameters in the model and loaded checkpoint should be identical.\n"
            "incorrect"
        )

        # model with the same likelihood but different parameters
        model_same_likelihood_other_prior = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            model_name="same_likelihood_other_prior",
            likelihood=GaussianLikelihood(),
        )
        with pytest.raises(ValueError) as error_msg:
            model_same_likelihood_other_prior.load_weights(
                model_path_manual, map_location="cpu"
            )
        assert str(error_msg.value).startswith(
            "The values of the hyper-parameters in the model and loaded checkpoint should be identical.\n"
            "incorrect"
        )

    def test_load_weights_params_check(self, tmpdir_fn):
        """
        Verify that the method comparing the parameters between the saved model and the loading model
        behave as expected, used to return meaningful error message instead of the torch.load ones.
        """
        model_name = "params_check"
        ckpt_path = os.path.join(tmpdir_fn, f"{model_name}.pt")
        # barebone model
        model = DLinearModel(
            input_chunk_length=4, output_chunk_length=1, n_epochs=1, **tfm_kwargs
        )
        model.fit(self.series[:10])
        model.save(ckpt_path)

        # identical model
        loading_model = DLinearModel(
            input_chunk_length=4, output_chunk_length=1, **tfm_kwargs
        )
        loading_model.load_weights(ckpt_path)

        # different optimizer
        loading_model = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=1,
            optimizer_cls=torch.optim.AdamW,
            **tfm_kwargs,
        )
        loading_model.load_weights(ckpt_path)

        model_summary_kwargs = {
            "pl_trainer_kwargs": dict(
                {"enable_model_sumamry": False}, **tfm_kwargs["pl_trainer_kwargs"]
            )
        }
        # different pl_trainer_kwargs
        loading_model = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=1,
            **model_summary_kwargs,
        )
        loading_model.load_weights(ckpt_path)

        # different input_chunk_length (tfm parameter)
        loading_model = DLinearModel(
            input_chunk_length=4 + 1, output_chunk_length=1, **tfm_kwargs
        )
        with pytest.raises(ValueError) as error_msg:
            loading_model.load_weights(ckpt_path)
        assert str(error_msg.value).startswith(
            "The values of the hyper-parameters in the model and loaded checkpoint should be identical.\n"
            "incorrect"
        )

        # different kernel size (cls specific parameter)
        loading_model = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=1,
            kernel_size=10,
            **tfm_kwargs,
        )
        with pytest.raises(ValueError) as error_msg:
            loading_model.load_weights(ckpt_path)
        assert str(error_msg.value).startswith(
            "The values of the hyper-parameters in the model and loaded checkpoint should be identical.\n"
            "incorrect"
        )

    def test_create_instance_new_model_no_name_set(self, tmpdir_fn):
        RNNModel(12, "RNN", 10, 10, work_dir=tmpdir_fn, **tfm_kwargs)
        # no exception is raised

    def test_create_instance_existing_model_with_name_no_fit(self, tmpdir_fn):
        model_name = "test_model"
        RNNModel(
            12,
            "RNN",
            10,
            10,
            work_dir=tmpdir_fn,
            model_name=model_name,
            **tfm_kwargs,
        )
        # no exception is raised

    @patch(
        "darts.models.forecasting.torch_forecasting_model.TorchForecastingModel.reset_model"
    )
    def test_create_instance_existing_model_with_name_force(
        self, patch_reset_model, tmpdir_fn
    ):
        model_name = "test_model"
        RNNModel(
            12,
            "RNN",
            10,
            10,
            work_dir=tmpdir_fn,
            model_name=model_name,
            **tfm_kwargs,
        )
        # no exception is raised
        # since no fit, there is no data stored for the model, hence `force_reset` does noting

        RNNModel(
            12,
            "RNN",
            10,
            10,
            work_dir=tmpdir_fn,
            model_name=model_name,
            force_reset=True,
            **tfm_kwargs,
        )
        patch_reset_model.assert_not_called()

    @patch(
        "darts.models.forecasting.torch_forecasting_model.TorchForecastingModel.reset_model"
    )
    def test_create_instance_existing_model_with_name_force_fit_with_reset(
        self, patch_reset_model, tmpdir_fn
    ):
        model_name = "test_model"
        model1 = RNNModel(
            12,
            "RNN",
            10,
            10,
            work_dir=tmpdir_fn,
            model_name=model_name,
            save_checkpoints=True,
            **tfm_kwargs,
        )
        # no exception is raised

        model1.fit(self.series, epochs=1)

        RNNModel(
            12,
            "RNN",
            10,
            10,
            work_dir=tmpdir_fn,
            model_name=model_name,
            save_checkpoints=True,
            force_reset=True,
            **tfm_kwargs,
        )
        patch_reset_model.assert_called_once()

    # TODO for PTL: currently we (have to (?)) create a mew PTL trainer object every time fit() is called which
    #  resets some of the model's attributes such as epoch and step counts. We have check whether there is another
    #  way of doing this.

    # n_epochs=20, fit|epochs=None, epochs_trained=0 - train for 20 epochs
    def test_train_from_0_n_epochs_20_no_fit_epochs(self):
        model1 = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=20,
            **tfm_kwargs,
        )

        model1.fit(self.series)

        assert 20 == model1.epochs_trained

    # n_epochs = 20, fit|epochs=None, epochs_trained=20 - train for another 20 epochs
    def test_train_from_20_n_epochs_40_no_fit_epochs(self):
        model1 = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=20,
            **tfm_kwargs,
        )

        model1.fit(self.series)
        assert 20 == model1.epochs_trained

        model1.fit(self.series)
        assert 20 == model1.epochs_trained

    # n_epochs = 20, fit|epochs=None, epochs_trained=10 - train for another 20 epochs
    def test_train_from_10_n_epochs_20_no_fit_epochs(self):
        model1 = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=20,
            **tfm_kwargs,
        )

        # simulate the case that user interrupted training with Ctrl-C after 10 epochs
        model1.fit(self.series, epochs=10)
        assert 10 == model1.epochs_trained

        model1.fit(self.series)
        assert 20 == model1.epochs_trained

    # n_epochs = 20, fit|epochs=15, epochs_trained=10 - train for 15 epochs
    def test_train_from_10_n_epochs_20_fit_15_epochs(self):
        model1 = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=20,
            **tfm_kwargs,
        )

        # simulate the case that user interrupted training with Ctrl-C after 10 epochs
        model1.fit(self.series, epochs=10)
        assert 10 == model1.epochs_trained

        model1.fit(self.series, epochs=15)
        assert 15 == model1.epochs_trained

    def test_load_weights_from_checkpoint(self, tmpdir_fn):
        ts_training, ts_test = self.series.split_before(90)
        original_model_name = "original"
        retrained_model_name = "retrained"
        # original model, checkpoints are saved
        model = RNNModel(
            12,
            "RNN",
            5,
            1,
            n_epochs=5,
            work_dir=tmpdir_fn,
            save_checkpoints=True,
            model_name=original_model_name,
            random_state=1,
            **tfm_kwargs,
        )
        model.fit(ts_training)
        original_preds = model.predict(10)
        original_mape = mape(original_preds, ts_test)

        # load last checkpoint of original model, train it for 2 additional epochs
        model_rt = RNNModel(
            12,
            "RNN",
            5,
            1,
            n_epochs=5,
            work_dir=tmpdir_fn,
            model_name=retrained_model_name,
            random_state=1,
            **tfm_kwargs,
        )
        model_rt.load_weights_from_checkpoint(
            model_name=original_model_name,
            work_dir=tmpdir_fn,
            best=False,
            map_location="cpu",
        )

        # must indicate series otherwise self.training_series must be saved in checkpoint
        loaded_preds = model_rt.predict(10, ts_training)
        # save/load checkpoint should produce identical predictions
        assert original_preds == loaded_preds

        model_rt.fit(ts_training)
        retrained_preds = model_rt.predict(10)
        retrained_mape = mape(retrained_preds, ts_test)
        assert retrained_mape < original_mape, (
            f"Retrained model has a greater error (mape) than the original model, "
            f"respectively {retrained_mape} and {original_mape}"
        )

        # raise Exception when trying to load ckpt weights in different architecture
        with pytest.raises(ValueError):
            model_rt = RNNModel(
                12,
                "RNN",
                10,  # loaded model has only 5 hidden_layers
                5,
            )
            model_rt.load_weights_from_checkpoint(
                model_name=original_model_name,
                work_dir=tmpdir_fn,
                best=False,
                map_location="cpu",
            )

        # raise Exception when trying to pass `weights_only`=True to `torch.load()`
        with pytest.raises(ValueError):
            model_rt = RNNModel(12, "RNN", 5, 5, **tfm_kwargs)
            model_rt.load_weights_from_checkpoint(
                model_name=original_model_name,
                work_dir=tmpdir_fn,
                best=False,
                weights_only=True,
                map_location="cpu",
            )

    def test_load_weights(self, tmpdir_fn):
        ts_training, ts_test = self.series.split_before(90)
        original_model_name = "original"
        retrained_model_name = "retrained"
        # original model, checkpoints are saved
        model = RNNModel(
            12,
            "RNN",
            5,
            1,
            n_epochs=5,
            work_dir=tmpdir_fn,
            save_checkpoints=False,
            model_name=original_model_name,
            random_state=1,
            **tfm_kwargs,
        )
        model.fit(ts_training)
        path_manual_save = os.path.join(tmpdir_fn, "RNN_manual_save.pt")
        model.save(path_manual_save)
        original_preds = model.predict(10)
        original_mape = mape(original_preds, ts_test)

        # load last checkpoint of original model, train it for 2 additional epochs
        model_rt = RNNModel(
            12,
            "RNN",
            5,
            1,
            n_epochs=5,
            work_dir=tmpdir_fn,
            model_name=retrained_model_name,
            random_state=1,
            **tfm_kwargs,
        )
        model_rt.load_weights(path=path_manual_save, map_location="cpu")

        # must indicate series otherwise self.training_series must be saved in checkpoint
        loaded_preds = model_rt.predict(10, ts_training)
        # save/load checkpoint should produce identical predictions
        assert original_preds == loaded_preds

        model_rt.fit(ts_training)
        retrained_preds = model_rt.predict(10)
        retrained_mape = mape(retrained_preds, ts_test)
        assert retrained_mape < original_mape, (
            f"Retrained model has a greater mape error than the original model, "
            f"respectively {retrained_mape} and {original_mape}"
        )

    def test_load_weights_with_float32_dtype(self, tmpdir_fn):
        ts_float32 = self.series.astype("float32")
        model_name = "test_model"
        ckpt_path = os.path.join(tmpdir_fn, f"{model_name}.pt")
        # barebone model
        model = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=1,
            n_epochs=1,
        )
        model.fit(ts_float32)
        model.save(ckpt_path)
        assert model.model._dtype == torch.float32  # type: ignore

        # identical model
        loading_model = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=1,
        )
        loading_model.load_weights(ckpt_path)
        loading_model.fit(ts_float32)
        assert loading_model.model._dtype == torch.float32  # type: ignore

    def test_multi_steps_pipeline(self, tmpdir_fn):
        ts_training, ts_val = self.series.split_before(75)
        pretrain_model_name = "pre-train"
        retrained_model_name = "re-train"

        # pretraining
        model = self.helper_create_RNNModel(pretrain_model_name, tmpdir_fn)
        model.fit(
            ts_training,
            val_series=ts_val,
        )

        # finetuning
        model = self.helper_create_RNNModel(retrained_model_name, tmpdir_fn)
        model.load_weights_from_checkpoint(
            model_name=pretrain_model_name,
            work_dir=tmpdir_fn,
            best=True,
            map_location="cpu",
        )
        model.fit(
            ts_training,
            val_series=ts_val,
        )

        # prediction
        model = model.load_from_checkpoint(
            model_name=retrained_model_name,
            work_dir=tmpdir_fn,
            best=True,
            map_location="cpu",
        )
        model.predict(4, series=ts_training)

    def test_load_from_checkpoint_w_custom_loss(self, tmpdir_fn):
        model_name = "pretraining_custom_loss"
        # model with a custom loss
        model = RNNModel(
            12,
            "RNN",
            5,
            1,
            n_epochs=1,
            work_dir=tmpdir_fn,
            model_name=model_name,
            save_checkpoints=True,
            force_reset=True,
            loss_fn=torch.nn.L1Loss(),
            **tfm_kwargs,
        )
        model.fit(self.series)

        loaded_model = RNNModel.load_from_checkpoint(
            model_name, tmpdir_fn, best=False, map_location="cpu"
        )
        # custom loss function should be properly restored from ckpt
        loss_fn_attrs = ["criterion", "train_criterion", "val_criterion"]
        for attr in loss_fn_attrs:
            assert isinstance(getattr(loaded_model.model, attr), torch.nn.L1Loss)

        loaded_model.fit(self.series, epochs=2)
        # calling fit() should not impact the loss function
        for attr in loss_fn_attrs:
            assert isinstance(getattr(loaded_model.model, attr), torch.nn.L1Loss)

    def test_load_from_checkpoint_w_metrics(self, tmpdir_fn):
        model_name = "pretraining_metrics"
        # model with one torch_metrics
        pl_trainer_kwargs = dict(
            {"logger": DummyLogger(), "log_every_n_steps": 1},
            **tfm_kwargs["pl_trainer_kwargs"],
        )
        model = RNNModel(
            12,
            "RNN",
            5,
            1,
            n_epochs=1,
            work_dir=tmpdir_fn,
            model_name=model_name,
            save_checkpoints=True,
            force_reset=True,
            torch_metrics=MeanAbsolutePercentageError(),
            pl_trainer_kwargs=pl_trainer_kwargs,
        )
        model.fit(self.series)
        # check train_metrics before loading
        assert isinstance(model.model.train_metrics, MetricCollection)
        assert len(model.model.train_metrics) == 1

        loaded_model = RNNModel.load_from_checkpoint(
            model_name,
            tmpdir_fn,
            best=False,
            map_location="cpu",
        )
        # custom loss function should be properly restored from ckpt torchmetrics.Metric
        assert isinstance(loaded_model.model.train_metrics, MetricCollection)
        assert len(loaded_model.model.train_metrics) == 1

    def test_optimizers(self):
        optimizers = [
            (torch.optim.Adam, {"lr": 0.001}),
            (torch.optim.SGD, {"lr": 0.001}),
        ]

        for optim_cls, optim_kwargs in optimizers:
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                optimizer_cls=optim_cls,
                optimizer_kwargs=optim_kwargs,
                **tfm_kwargs,
            )
            # should not raise an error
            model.fit(self.series, epochs=1)

    @pytest.mark.parametrize(
        "lr_scheduler",
        [
            (torch.optim.lr_scheduler.StepLR, {"step_size": 10}),
            (
                torch.optim.lr_scheduler.ReduceLROnPlateau,
                {
                    "threshold": 0.001,
                    "monitor": "train_loss",
                    "interval": "step",
                    "frequency": 2,
                },
            ),
            (torch.optim.lr_scheduler.ExponentialLR, {"gamma": 0.09}),
        ],
    )
    def test_lr_schedulers(self, lr_scheduler):
        lr_scheduler_cls, lr_scheduler_kwargs = lr_scheduler
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            lr_scheduler_cls=lr_scheduler_cls,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            **tfm_kwargs,
        )
        # should not raise an error
        model.fit(self.series, epochs=1)

    def test_wrong_model_creation_params(self):
        valid_kwarg = {"pl_trainer_kwargs": {}}
        invalid_kwarg = {"some_invalid_kwarg": None}

        # valid params should not raise an error
        _ = RNNModel(12, "RNN", 10, 10, **valid_kwarg)

        # invalid params should raise an error
        with pytest.raises(ValueError):
            _ = RNNModel(12, "RNN", 10, 10, **invalid_kwarg)

    def test_metrics(self):
        metric = MeanAbsolutePercentageError()
        metric_collection = MetricCollection([
            MeanAbsolutePercentageError(),
            MeanAbsoluteError(),
        ])

        model_kwargs = {
            "logger": DummyLogger(),
            "log_every_n_steps": 1,
            **tfm_kwargs["pl_trainer_kwargs"],
        }
        # test single metric
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            torch_metrics=metric,
            pl_trainer_kwargs=model_kwargs,
        )
        model.fit(self.series)

        # test metric collection
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            torch_metrics=metric_collection,
            pl_trainer_kwargs=model_kwargs,
        )
        model.fit(self.series)

        # test multivariate series
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            torch_metrics=metric_collection,
            pl_trainer_kwargs=model_kwargs,
        )
        model.fit(self.multivariate_series)

    def test_metrics_w_likelihood(self):
        metric = MeanAbsolutePercentageError()
        metric_collection = MetricCollection([
            MeanAbsolutePercentageError(),
            MeanAbsoluteError(),
        ])
        model_kwargs = {
            "logger": DummyLogger(),
            "log_every_n_steps": 1,
            **tfm_kwargs["pl_trainer_kwargs"],
        }
        # test single metric
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            likelihood=GaussianLikelihood(),
            torch_metrics=metric,
            pl_trainer_kwargs=model_kwargs,
        )
        model.fit(self.series)

        # test metric collection
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            likelihood=GaussianLikelihood(),
            torch_metrics=metric_collection,
            pl_trainer_kwargs=model_kwargs,
        )
        model.fit(self.series)

        # test multivariate series
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            likelihood=GaussianLikelihood(),
            torch_metrics=metric_collection,
            pl_trainer_kwargs=model_kwargs,
        )
        model.fit(self.multivariate_series)

    def test_invalid_metrics(self):
        torch_metrics = ["invalid"]
        with pytest.raises(AttributeError):
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                n_epochs=1,
                torch_metrics=torch_metrics,
                **tfm_kwargs,
            )
            model.fit(self.series)

    def test_stateful_metrics(self):
        torch_metrics = NumsCalled()
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            n_epochs=1,
            torch_metrics=torch_metrics,
            **tfm_kwargs,
        )
        model.fit(self.series)
        assert model.model.trainer.logged_metrics["train_NumsCalled"] > 1

    @pytest.mark.slow
    def test_lr_find(self):
        train_series, val_series = self.series[:-40], self.series[-40:]
        model = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)
        # find the learning rate
        res = model.lr_find(series=train_series, val_series=val_series, epochs=50)
        assert isinstance(res, _LRFinder)
        assert res.suggestion() is not None
        # verify that learning rate finder bypasses the `fit` logic
        assert model.model is None
        assert not model._fit_called
        # cannot predict with an untrained model
        with pytest.raises(ValueError):
            model.predict(n=3, series=self.series)

        # check that results are reproducible
        model = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)
        res2 = model.lr_find(series=train_series, val_series=val_series, epochs=50)
        assert res.suggestion() == res2.suggestion()

        # check that suggested learning rate is better than the worst
        lr_worst = res.results["lr"][np.argmax(res.results["loss"])]
        lr_suggested = res.suggestion()
        scores = {}
        for lr, lr_name in zip([lr_worst, lr_suggested], ["worst", "suggested"]):
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                n_epochs=10,
                random_state=42,
                optimizer_cls=torch.optim.Adam,
                optimizer_kwargs={"lr": lr},
                **tfm_kwargs,
            )
            model.fit(train_series)
            scores[lr_name] = mape(
                val_series, model.predict(len(val_series), series=train_series)
            )
        assert scores["worst"] > scores["suggested"]

    def test_encoders(self, tmpdir_fn):
        series = tg.linear_timeseries(length=10)
        pc = tg.linear_timeseries(length=12)
        fc = tg.linear_timeseries(length=13)
        # 1 == output_chunk_length, 3 > output_chunk_length
        ns = [1, 3]

        model = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            add_encoders={
                "datetime_attribute": {"past": ["hour"], "future": ["month"]}
            },
        )
        model.fit(series)
        for n in ns:
            _ = model.predict(n=n)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, past_covariates=pc)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, future_covariates=fc)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, past_covariates=pc, future_covariates=fc)

        model = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            add_encoders={
                "datetime_attribute": {"past": ["hour"], "future": ["month"]}
            },
        )
        for n in ns:
            model.fit(series, past_covariates=pc)
            _ = model.predict(n=n)
            _ = model.predict(n=n, past_covariates=pc)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, future_covariates=fc)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, past_covariates=pc, future_covariates=fc)

        model = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            add_encoders={
                "datetime_attribute": {"past": ["hour"], "future": ["month"]}
            },
        )
        for n in ns:
            model.fit(series, future_covariates=fc)
            _ = model.predict(n=n)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, past_covariates=pc)
            _ = model.predict(n=n, future_covariates=fc)
            with pytest.raises(ValueError):
                _ = model.predict(n=n, past_covariates=pc, future_covariates=fc)

        model = self.helper_create_DLinearModel(
            work_dir=tmpdir_fn,
            add_encoders={
                "datetime_attribute": {"past": ["hour"], "future": ["month"]}
            },
        )
        for n in ns:
            model.fit(series, past_covariates=pc, future_covariates=fc)
            _ = model.predict(n=n)
            _ = model.predict(n=n, past_covariates=pc)
            _ = model.predict(n=n, future_covariates=fc)
            _ = model.predict(n=n, past_covariates=pc, future_covariates=fc)

    @pytest.mark.parametrize("model_config", models)
    def test_val_set(self, model_config):
        """Test whether these evaluation set parameters are passed to the PyTorch Lightning Trainer"""
        with patch("pytorch_lightning.Trainer.fit") as fit_patch:
            self.helper_check_val_set(*model_config, fit_patch)

    def test_dataloader_kwargs_setup(self):
        train_series, val_series = self.series[:-40], self.series[-40:]
        model = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)

        with patch("pytorch_lightning.Trainer.fit") as fit_patch:
            model.fit(train_series, val_series=val_series)
            assert "train_dataloaders" in fit_patch.call_args.kwargs
            assert "val_dataloaders" in fit_patch.call_args.kwargs

            train_dl = fit_patch.call_args.kwargs["train_dataloaders"]
            assert isinstance(train_dl, DataLoader)
            val_dl = fit_patch.call_args.kwargs["val_dataloaders"]
            assert isinstance(val_dl, DataLoader)

            dl_defaults = {
                "batch_size": model.batch_size,
                "pin_memory": True,
                "drop_last": False,
                "collate_fn": model._batch_collate_fn,
            }
            assert all([getattr(train_dl, k) == v for k, v in dl_defaults.items()])
            # shuffle=True gives random sampler
            assert isinstance(train_dl.sampler, RandomSampler)

            assert all([getattr(val_dl, k) == v for k, v in dl_defaults.items()])
            # shuffle=False gives sequential sampler
            assert isinstance(val_dl.sampler, SequentialSampler)

            # check that overwriting the dataloader kwargs works
            dl_custom = dict(dl_defaults, **{"batch_size": 50, "drop_last": True})
            model.fit(train_series, val_series=val_series, dataloader_kwargs=dl_custom)
            train_dl = fit_patch.call_args.kwargs["train_dataloaders"]
            val_dl = fit_patch.call_args.kwargs["val_dataloaders"]
            assert all([getattr(train_dl, k) == v for k, v in dl_custom.items()])
            assert all([getattr(val_dl, k) == v for k, v in dl_custom.items()])

        with patch("pytorch_lightning.Trainer.predict") as pred_patch:
            # calling predict with the patch will raise an error, but we only need to
            # check the dataloader setup
            with pytest.raises(Exception):
                model.predict(n=1)
            assert "dataloaders" in pred_patch.call_args.kwargs
            pred_dl = pred_patch.call_args.kwargs["dataloaders"]
            assert isinstance(pred_dl, DataLoader)
            assert all([getattr(pred_dl, k) == v for k, v in dl_defaults.items()])
            # shuffle=False gives sequential sampler
            assert isinstance(val_dl.sampler, SequentialSampler)

            # check that overwriting the dataloader kwargs works
            with pytest.raises(Exception):
                model.predict(n=1, dataloader_kwargs=dl_custom)
            pred_dl = pred_patch.call_args.kwargs["dataloaders"]
            assert all([getattr(pred_dl, k) == v for k, v in dl_custom.items()])

    def test_dataloader_kwargs_fit_predict(self):
        train_series, val_series = self.series[:-40], self.series[-40:]
        model = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)

        model.fit(
            train_series,
            val_series=val_series,
            dataloader_kwargs={"batch_size": 100, "shuffle": False},
        )

        # check same results with default batch size (32) and custom batch size
        preds_default = model.predict(
            n=2,
            series=[train_series, val_series],
        )
        preds_custom = model.predict(
            n=2,
            series=[train_series, val_series],
            dataloader_kwargs={"batch_size": 100},
        )
        assert preds_default == preds_custom

    def helper_check_val_set(self, model_cls, model_kwargs, fit_patch):
        # naive models don't call the Trainer
        if issubclass(model_cls, _GlobalNaiveModel):
            return

        series1 = tg.sine_timeseries(length=11, column_name="tg_1")
        series2 = tg.sine_timeseries(length=11, column_name="tg_2") / 2 + 10
        series = series1.stack(series2)
        series = series.with_static_covariates(
            pd.DataFrame({"sc1": [0, 1], "sc2": [3, 4]})
        )
        pc = series1 * 10 - 3
        fc = TimeSeries.from_times_and_values(
            times=series.time_index, values=series.values() * -1, columns=["fc1", "fc2"]
        )
        model = model_cls(**model_kwargs)

        # check that an error is raised with an invalid validation series
        fit_kwargs = {
            "series": series,
            "val_series": series["tg_1"],
        }
        invalid_series_txt = "`series`"
        if model.supports_past_covariates:
            fit_kwargs["past_covariates"] = pc
            fit_kwargs["val_past_covariates"] = pc
        if model.supports_future_covariates:
            fit_kwargs["future_covariates"] = fc
            fit_kwargs["val_future_covariates"] = fc["fc1"]
            invalid_series_txt += ", `future_covariates`"
        if model.supports_static_covariates:
            invalid_series_txt += ", `static_covariates`"

        with pytest.raises(ValueError) as err:
            model.fit(**fit_kwargs)
        msg_expected = (
            f"The dimensions of the ({invalid_series_txt}) between "
            "the training and validation set do not match."
        )
        assert str(err.value) == msg_expected

        # check that an error is raised if only second validation series are invalid
        fit_kwargs = {
            "series": series,
            "val_series": [series, series["tg_1"]],
        }
        invalid_series_txt = "`series`"
        if model.supports_past_covariates:
            fit_kwargs["past_covariates"] = pc
            fit_kwargs["val_past_covariates"] = [pc, pc]
        if model.supports_future_covariates:
            fit_kwargs["future_covariates"] = fc
            fit_kwargs["val_future_covariates"] = [fc, fc["fc1"]]
            invalid_series_txt += ", `future_covariates`"
        if model.supports_static_covariates:
            invalid_series_txt += ", `static_covariates`"

        with pytest.raises(ValueError) as err:
            model.fit(**fit_kwargs)
        msg_expected = (
            f"The dimensions of the ({invalid_series_txt}) between "
            "the training and validation set at sequence/list index `1` do not match."
        )
        assert str(err.value) == msg_expected

        fit_kwargs = {"series": series, "val_series": series}
        if model.supports_past_covariates:
            fit_kwargs["past_covariates"] = pc
            fit_kwargs["val_past_covariates"] = pc
        if model.supports_future_covariates:
            fit_kwargs["future_covariates"] = fc
            fit_kwargs["val_future_covariates"] = fc

        model.fit(**fit_kwargs)
        # fit called only once
        assert fit_patch.call_count == 1

        train_ds = fit_patch.call_args[1]["train_dataloaders"].dataset
        val_dl = fit_patch.call_args[1]["val_dataloaders"]
        assert val_dl is not None
        val_ds = val_dl.dataset

        # check same dataset type
        assert isinstance(val_ds, train_ds.__class__)

        # check that input in first batch have same dimensions
        train_sample = train_ds[0]
        val_sample = val_ds[0]
        assert len(val_sample) == len(train_sample)
        for x_train, x_val in zip(train_sample, val_sample):
            if x_train is None:
                assert x_val is None
            else:
                assert x_val.shape[1:] == x_train.shape[1:]

    @pytest.mark.parametrize("model_config", models)
    def test_rin(self, model_config):
        model_cls, kwargs = model_config
        model_no_rin = model_cls(use_reversible_instance_norm=False, **kwargs)
        model_rin = model_cls(use_reversible_instance_norm=True, **kwargs)

        # univariate no RIN
        model_no_rin.fit(self.series)
        assert not model_no_rin.model.use_reversible_instance_norm
        assert model_no_rin.model.rin is None

        # univariate with RIN
        model_rin.fit(self.series)
        if issubclass(model_cls, RNNModel):
            # RNNModel will not use RIN
            assert not model_rin.model.use_reversible_instance_norm
            assert model_rin.model.rin is None
            return
        else:
            assert model_rin.model.use_reversible_instance_norm
            assert isinstance(model_rin.model.rin, RINorm)
            assert model_rin.model.rin.input_dim == self.series.n_components
        # multivariate with RIN
        model_rin_mv = model_rin.untrained_model()
        model_rin_mv.fit(self.multivariate_series)
        assert model_rin_mv.model.use_reversible_instance_norm
        assert isinstance(model_rin_mv.model.rin, RINorm)
        assert model_rin_mv.model.rin.input_dim == self.multivariate_series.n_components

    @pytest.mark.parametrize("use_mc_dropout", [False, True])
    def test_mc_dropout_active(self, use_mc_dropout):
        """Test that model activates dropout ."""

        class CheckMCDropout(Callback):
            def __init__(self, activate_mc_dropout):
                self.use_mc_dropout = activate_mc_dropout

            @staticmethod
            def _check_dropout_activity(pl_module, expected_active: bool):
                dropouts = pl_module._get_mc_dropout_modules()
                assert all([
                    dropout.mc_dropout_enabled is expected_active
                    for dropout in dropouts
                ])

            def on_train_batch_start(self, *args, **kwargs) -> None:
                self._check_dropout_activity(args[1], expected_active=True)

            def on_validation_batch_start(self, *args, **kwargs) -> None:
                self._check_dropout_activity(args[1], expected_active=False)

            def on_predict_batch_start(self, *args, **kwargs) -> None:
                self._check_dropout_activity(
                    args[1], expected_active=self.use_mc_dropout
                )

        series = self.series[:20]
        pl_trainer_kwargs = copy.deepcopy(tfm_kwargs)
        pl_trainer_kwargs["pl_trainer_kwargs"]["callbacks"] = [
            CheckMCDropout(activate_mc_dropout=use_mc_dropout)
        ]
        model = TiDEModel(10, 10, dropout=0.1, random_state=42, **pl_trainer_kwargs)
        model.fit(series, val_series=series, epochs=1)

        num_samples = 1 if not use_mc_dropout else 10
        preds = model.predict(
            n=10, series=series, mc_dropout=use_mc_dropout, num_samples=num_samples
        )
        assert preds.n_samples == num_samples

    @pytest.mark.parametrize("use_mc_dropout", [False, True])
    def test_dropout_output(self, use_mc_dropout):
        """Test that model without dropout generates different results than one which uses near-full dropout."""
        series = self.series[:20]
        num_samples = 1 if not use_mc_dropout else 10

        # dropouts for overfit and underfit
        preds = []
        for dropout in [0.0, 0.99]:
            model = TiDEModel(10, 10, dropout=dropout, random_state=42, **tfm_kwargs)
            model.fit(series, val_series=series, epochs=1)
            preds.append(
                model.predict(
                    n=10,
                    series=series,
                    mc_dropout=use_mc_dropout,
                    num_samples=num_samples,
                ).all_values()
            )
        assert not np.array_equal(preds[0], preds[1])

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            models,
            [3, 7, 10],
        ),
    )
    def test_output_shift(self, config):
        """Tests shifted output for shift smaller than, equal to, and larger than output_chunk_length.
        RNNModel does not support shift output chunk.
        """
        np.random.seed(0)
        (model_cls, model_kwargs), shift = config
        if issubclass(model_cls, RNNModel):
            return

        model_kwargs = copy.deepcopy(model_kwargs)
        model_kwargs.pop("input_chunk_length")
        model_kwargs.pop("output_chunk_length")

        if issubclass(model_cls, TFTModel):
            model_kwargs.update({"likelihood": None, "loss_fn": torch.nn.MSELoss()})

        icl = 8
        ocl = 7
        series = tg.gaussian_timeseries(
            length=28, start=pd.Timestamp("2000-01-01"), freq="d"
        )

        model = self.helper_create_torch_model(
            model_cls, icl, ocl, shift, **model_kwargs
        )
        model.fit(series)

        # no auto-regression with shifted output
        with pytest.raises(ValueError) as err:
            _ = model.predict(n=ocl + 1)
        assert str(err.value).startswith("Cannot perform auto-regression")

        # pred starts with a shift
        for ocl_test in [ocl - 1, ocl]:
            pred = model.predict(n=ocl_test)
            assert pred.start_time() == series.end_time() + (shift + 1) * series.freq
            assert len(pred) == ocl_test
            assert pred.freq == series.freq

        # check that shifted output chunk results with encoders are the
        # same as using identical covariates

        # model trained on encoders
        cov_support = []
        covs = {}
        if model.supports_past_covariates:
            cov_support.append("past")
            covs["past_covariates"] = tg.datetime_attribute_timeseries(
                series,
                attribute="dayofweek",
                add_length=0,
            )
        if model.supports_future_covariates:
            cov_support.append("future")
            covs["future_covariates"] = tg.datetime_attribute_timeseries(
                series,
                attribute="dayofweek",
                add_length=ocl + shift,
            )

        if not cov_support:
            return

        add_encoders = {
            "datetime_attribute": {cov: ["dayofweek"] for cov in cov_support}
        }
        model_enc_shift = self.helper_create_torch_model(
            model_cls, icl, ocl, shift, add_encoders=add_encoders, **model_kwargs
        )
        model_enc_shift.fit(series)

        # model trained with identical covariates
        model_fc_shift = self.helper_create_torch_model(
            model_cls, icl, ocl, shift, **model_kwargs
        )

        model_fc_shift.fit(series, **covs)

        pred_enc = model_enc_shift.predict(n=ocl)
        pred_fc = model_fc_shift.predict(n=ocl)
        assert pred_enc == pred_fc

        # check that historical forecasts works properly
        hist_fc_start = -(ocl + shift)
        pred_last_hist_fc = model_fc_shift.predict(n=ocl, series=series[:hist_fc_start])
        # non-optimized hist fc
        hist_fc = model_fc_shift.historical_forecasts(
            series=series,
            start=hist_fc_start,
            start_format="position",
            retrain=False,
            forecast_horizon=ocl,
            last_points_only=False,
            enable_optimization=False,
            **covs,
        )
        assert len(hist_fc) == 1
        assert hist_fc[0] == pred_last_hist_fc
        # optimized hist fc, due to batch predictions, slight deviations in values
        hist_fc_opt = model_fc_shift.historical_forecasts(
            series=series,
            start=hist_fc_start,
            start_format="position",
            retrain=False,
            forecast_horizon=ocl,
            last_points_only=False,
            enable_optimization=True,
            **covs,
        )
        assert len(hist_fc_opt) == 1
        assert hist_fc_opt[0].time_index.equals(pred_last_hist_fc.time_index)
        np.testing.assert_array_almost_equal(
            hist_fc_opt[0].values(copy=False), pred_last_hist_fc.values(copy=False)
        )

        # covs too short
        for cov_name in cov_support:
            with pytest.raises(ValueError) as err:
                add_covs = {
                    cov_name + "_covariates": covs[cov_name + "_covariates"][:-1]
                }
                _ = model_fc_shift.predict(n=ocl, **add_covs)
            assert f"provided {cov_name} covariates at dataset index" in str(err.value)

    @pytest.mark.parametrize("config", itertools.product(models, [2, 3, 4]))
    def test_multi_ts_prediction(self, config):
        (model_cls, model_kwargs), n = config
        model_kwargs = copy.deepcopy(model_kwargs)
        model_kwargs["output_chunk_length"] = 3
        series = tg.linear_timeseries(
            length=model_kwargs["input_chunk_length"]
            + model_kwargs["output_chunk_length"]
        )
        model = model_cls(**model_kwargs)
        model.fit(series)
        # test with more series that `n`
        n_series_more = 5
        pred = model.predict(n=n, series=[series] * n_series_more)
        assert len(pred) == n_series_more
        assert all(len(p) == n for p in pred)

        # test with less series that `n`
        n_series_less = 1
        pred = model.predict(n=n, series=[series] * n_series_less)
        assert len(pred) == n_series_less
        assert all(len(p) == n for p in pred)

    @pytest.mark.parametrize(
        "config",
        itertools.product(models, [True, False], [True, False], [True, False]),
    )
    def test_weights(self, config):
        (model_cls, model_kwargs), built_in_weight, single_series, univ_series = config
        model_kwargs = copy.deepcopy(model_kwargs)
        # take larger learning rate to make network weights updates more pronounced
        model_kwargs["optimizer_kwargs"] = {"lr": 0.1}
        model_kwargs["pl_trainer_kwargs"]["max_epochs"] = 2
        model_kwargs["pl_trainer_kwargs"]["fast_dev_run"] = False
        # create more than one batch sample as otherwise linear sample weight would always be `1.`
        ts = tg.linear_timeseries(
            length=model_kwargs["input_chunk_length"]
            + model_kwargs["output_chunk_length"]
            + 1
        )
        if not univ_series:
            ts = ts.stack(ts)

        if built_in_weight:
            weights = "linear"
        else:
            weights = np.expand_dims(np.linspace(0, 1, len(ts)), -1)
            if not univ_series:
                weights = np.concatenate([weights] * ts.n_components, axis=1)
            weights = ts.with_values(weights)

        if not single_series:
            ts = [ts] * 2
            weights = weights if built_in_weight else [weights] * 2

        model = model_cls(**model_kwargs)
        model.fit(ts, sample_weight=weights)
        preds = model.predict(n=3, series=ts)

        # check deterministic results
        model_identical = model_cls(**model_kwargs)
        model_identical.fit(ts, sample_weight=weights)
        preds_identical = model_identical.predict(n=3, series=ts)

        if single_series:
            preds = [preds]
            preds_identical = [preds_identical]

        for pred, preds_identical in zip(preds, preds_identical):
            np.testing.assert_array_almost_equal(
                pred.all_values(), preds_identical.all_values()
            )

        model_no_weight = model_cls(**model_kwargs)
        model_no_weight.fit(ts, sample_weight=None)
        preds_no_weight = model_no_weight.predict(n=3, series=ts)

        if single_series:
            preds_no_weight = [preds_no_weight]

        for pred, pred_no_weight in zip(preds, preds_no_weight):
            if isinstance(model, _GlobalNaiveModel):
                # naive models don't learn, so output should be the same
                np.testing.assert_array_almost_equal(
                    pred.all_values(), pred_no_weight.all_values()
                )
            else:
                # all other models should have different results from sample weights
                with pytest.raises(AssertionError):
                    np.testing.assert_array_almost_equal(
                        pred.all_values(), pred_no_weight.all_values()
                    )

        model_kwargs["pl_trainer_kwargs"]["max_epochs"] = 1
        model_kwargs["pl_trainer_kwargs"]["fast_dev_run"] = True
        model = model_cls(**model_kwargs)
        # try with validation series and only train weights
        model.fit(ts, val_series=ts, sample_weight=weights)

        # try with validation series and only val weights
        model.fit(ts, val_series=ts, val_sample_weight=weights)

        # try with validation series and train and val weights
        model.fit(ts, val_series=ts, sample_weight=weights, val_sample_weight=weights)

    def test_invalid_weights(self):
        model_cls, model_kwargs = models[0]
        ts = tg.linear_timeseries(
            length=model_kwargs["input_chunk_length"]
            + model_kwargs["output_chunk_length"]
        )

        # weights too short
        model = model_cls(**model_kwargs)
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight=ts[:-1])
        assert (
            str(err.value)
            == "Missing sample weights; could not find sample weights in index value range: "
            "2000-01-11 00:00:00 - 2000-01-11 00:00:00."
        )

        # same number of series
        model = model_cls(**model_kwargs)
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight=[ts, ts])
        assert (
            str(err.value)
            == "The provided sequence of target `series` must have the same length as the "
            "provided sequence of `sample_weight`."
        )

        # same number of components
        model = model_cls(**model_kwargs)
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight=ts.stack(ts))
        assert (
            str(err.value)
            == "The number of components in `sample_weight` must either be `1` or match the "
            "number of target series components `1`. (0-th series)"
        )
        # with correct number it works
        model = model_cls(**model_kwargs)
        model.fit(ts.stack(ts), sample_weight=ts.stack(ts))
        # or with multivar ts and single component weights (globally applied)
        model = model_cls(**model_kwargs)
        model.fit(ts.stack(ts), sample_weight=ts)

        # invalid string
        model = model_cls(**model_kwargs)
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight="invalid")
        assert str(err.value).startswith("Invalid `sample_weight` value: `'invalid'`. ")

    @pytest.mark.parametrize(
        "likelihood",
        [
            QuantileRegression([0.1, 0.5, 0.9]),
            LaplaceLikelihood(),
            GaussianLikelihood(),
            CauchyLikelihood(),
        ],
    )
    def test_weights_probabilistic(self, likelihood):
        model_cls, model_kwargs = models[0]
        ts = tg.linear_timeseries(
            length=model_kwargs["input_chunk_length"]
            + model_kwargs["output_chunk_length"]
        )

        model_kwargs = copy.deepcopy(model_kwargs)
        model_kwargs["likelihood"] = likelihood
        model_kwargs["loss_fn"] = None

        model = model_cls(**model_kwargs)
        model.fit(ts, sample_weight=ts)
        pred = model.predict(n=3, num_samples=10)

        # check results are deterministic with same sample weights
        model_same = model_cls(**model_kwargs)
        model_same.fit(ts, sample_weight=ts)
        pred_same = model_same.predict(n=3, num_samples=10)
        np.testing.assert_array_almost_equal(pred.all_values(), pred_same.all_values())

        # check different results without sample weights
        model_no_weight = model_cls(**model_kwargs)
        model_no_weight.fit(ts, sample_weight=ts)
        pred_no_weight = model.predict(n=3, num_samples=10)

        # all other models should have different results from sample weights
        with pytest.raises(AssertionError):
            np.testing.assert_array_almost_equal(
                pred.all_values(), pred_no_weight.all_values()
            )

    def helper_equality_encoders(
        self, first_encoders: dict[str, Any], second_encoders: dict[str, Any]
    ):
        if first_encoders is None:
            first_encoders = {}
        if second_encoders is None:
            second_encoders = {}
        assert {k: v for k, v in first_encoders.items() if k != "transformer"} == {
            k: v for k, v in second_encoders.items() if k != "transformer"
        }

    def helper_equality_encoders_transfo(
        self, first_encoders: dict[str, Any], second_encoders: dict[str, Any]
    ):
        if first_encoders is None:
            first_encoders = {}
        if second_encoders is None:
            second_encoders = {}
        assert (
            first_encoders.get("transformer", None).__class__
            == second_encoders.get("transformer", None).__class__
        )

    def helper_create_RNNModel(self, model_name: str, tmpdir_fn):
        return RNNModel(
            input_chunk_length=4,
            hidden_dim=3,
            add_encoders={
                "cyclic": {"past": ["month"]},
                "datetime_attribute": {
                    "past": ["hour"],
                },
                "transformer": Scaler(),
            },
            n_epochs=2,
            model_name=model_name,
            work_dir=tmpdir_fn,
            force_reset=True,
            save_checkpoints=True,
            **tfm_kwargs,
        )

    def helper_create_DLinearModel(
        self,
        work_dir: Optional[str] = None,
        model_name: str = "unitest_model",
        add_encoders: Optional[dict] = None,
        save_checkpoints: bool = False,
        likelihood: Optional[Likelihood] = None,
        output_chunk_length: int = 1,
        **kwargs,
    ):
        return DLinearModel(
            input_chunk_length=4,
            output_chunk_length=output_chunk_length,
            model_name=model_name,
            add_encoders=add_encoders,
            work_dir=work_dir,
            save_checkpoints=save_checkpoints,
            random_state=42,
            force_reset=True,
            n_epochs=1,
            likelihood=likelihood,
            **tfm_kwargs,
            **kwargs,
        )

    def helper_create_torch_model(self, model_cls, icl, ocl, shift, **kwargs):
        params = {
            "input_chunk_length": icl,
            "output_chunk_length": ocl,
            "output_chunk_shift": shift,
            "n_epochs": 1,
            "random_state": 42,
        }
        params.update(tfm_kwargs)
        params.update(kwargs)
        return model_cls(**params)



"""
Time-series Dense Encoder (TiDE)
------
"""

from typing import Optional

import torch
import torch.nn as nn

from darts.logging import get_logger, raise_log
from darts.models.forecasting.pl_forecasting_module import (
    PLMixedCovariatesModule,
    io_processor,
)
from darts.models.forecasting.torch_forecasting_model import MixedCovariatesTorchModel
from darts.utils.torch import MonteCarloDropout

MixedCovariatesTrainTensorType = tuple[
    torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor
]


logger = get_logger(__name__)


class _ResidualBlock(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_size: int,
        dropout: float,
        use_layer_norm: bool,
    ):
        """Pytorch module implementing the Residual Block from the TiDE paper."""
        super().__init__()

        # dense layer with ReLU activation with dropout
        self.dense = nn.Sequential(
            nn.Linear(input_dim, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, output_dim),
            MonteCarloDropout(dropout),
        )

        # linear skip connection from input to output of self.dense
        self.skip = nn.Linear(input_dim, output_dim)

        # layer normalization as output
        if use_layer_norm:
            self.layer_norm = nn.LayerNorm(output_dim)
        else:
            self.layer_norm = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # residual connection
        x = self.dense(x) + self.skip(x)

        # layer normalization
        if self.layer_norm is not None:
            x = self.layer_norm(x)

        return x


class _TideModule(PLMixedCovariatesModule):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        future_cov_dim: int,
        static_cov_dim: int,
        nr_params: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        decoder_output_dim: int,
        hidden_size: int,
        temporal_decoder_hidden: int,
        temporal_width_past: int,
        temporal_width_future: int,
        use_layer_norm: bool,
        dropout: float,
        temporal_hidden_size_past: Optional[int] = None,
        temporal_hidden_size_future: Optional[int] = None,
        **kwargs,
    ):
        """Pytorch module implementing the TiDE architecture.

        Parameters
        ----------
        input_dim
            The number of input components (target + optional past covariates + optional future covariates).
        output_dim
            Number of output components in the target.
        future_cov_dim
            Number of future covariates.
        static_cov_dim
            Number of static covariates.
        nr_params
            The number of parameters of the likelihood (or 1 if no likelihood is used).
        num_encoder_layers
            Number of stacked Residual Blocks in the encoder.
        num_decoder_layers
            Number of stacked Residual Blocks in the decoder.
        decoder_output_dim
            The number of output components of the decoder.
        hidden_size
            The width of the hidden layers in the encoder/decoder Residual Blocks.
        temporal_decoder_hidden
            The width of the hidden layers in the temporal decoder.
        temporal_width_past
            The width of the past covariate embedding space.
        temporal_width_future
            The width of the future covariate embedding space.
        temporal_hidden_size_past
            The width of the hidden layers in the past covariate projection Residual Block.
        temporal_hidden_size_future
            The width of the hidden layers in the future covariate projection Residual Block.
        use_layer_norm
            Whether to use layer normalization in the Residual Blocks.
        dropout
            Dropout probability
        **kwargs
            all parameters required for :class:`darts.models.forecasting.pl_forecasting_module.PLForecastingModule`
            base class.

        Inputs
        ------
        x
            Tuple of Tensors `(x_past, x_future, x_static)` where `x_past` is the input/past chunk and
            `x_future`is the output/future chunk. Input dimensions are `(batch_size, time_steps, components)`
        Outputs
        -------
        y
            Tensor of shape `(batch_size, output_chunk_length, output_dim, nr_params)`

        """

        super().__init__(**kwargs)

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.past_cov_dim = input_dim - output_dim - future_cov_dim
        self.future_cov_dim = future_cov_dim
        self.static_cov_dim = static_cov_dim
        self.nr_params = nr_params
        self.num_encoder_layers = num_encoder_layers
        self.num_decoder_layers = num_decoder_layers
        self.decoder_output_dim = decoder_output_dim
        self.hidden_size = hidden_size
        self.temporal_decoder_hidden = temporal_decoder_hidden
        self.use_layer_norm = use_layer_norm
        self.dropout = dropout
        self.temporal_width_past = temporal_width_past
        self.temporal_width_future = temporal_width_future
        self.temporal_hidden_size_past = temporal_hidden_size_past or hidden_size
        self.temporal_hidden_size_future = temporal_hidden_size_future or hidden_size

        # past covariates handling: either feature projection, raw features, or no features
        self.past_cov_projection = None
        if self.past_cov_dim and temporal_width_past:
            # residual block for past covariates feature projection
            self.past_cov_projection = _ResidualBlock(
                input_dim=self.past_cov_dim,
                output_dim=temporal_width_past,
                hidden_size=temporal_hidden_size_past,
                use_layer_norm=use_layer_norm,
                dropout=dropout,
            )
            past_covariates_flat_dim = self.input_chunk_length * temporal_width_past
        elif self.past_cov_dim:
            # skip projection and use raw features
            past_covariates_flat_dim = self.input_chunk_length * self.past_cov_dim
        else:
            past_covariates_flat_dim = 0

        # future covariates handling: either feature projection, raw features, or no features
        self.future_cov_projection = None
        if future_cov_dim and self.temporal_width_future:
            # residual block for future covariates feature projection
            self.future_cov_projection = _ResidualBlock(
                input_dim=future_cov_dim,
                output_dim=temporal_width_future,
                hidden_size=temporal_hidden_size_future,
                use_layer_norm=use_layer_norm,
                dropout=dropout,
            )
            historical_future_covariates_flat_dim = (
                self.input_chunk_length + self.output_chunk_length
            ) * temporal_width_future
        elif future_cov_dim:
            # skip projection and use raw features
            historical_future_covariates_flat_dim = (
                self.input_chunk_length + self.output_chunk_length
            ) * future_cov_dim
        else:
            historical_future_covariates_flat_dim = 0

        encoder_dim = (
            self.input_chunk_length * output_dim
            + past_covariates_flat_dim
            + historical_future_covariates_flat_dim
            + static_cov_dim
        )

        self.encoders = nn.Sequential(
            _ResidualBlock(
                input_dim=encoder_dim,
                output_dim=hidden_size,
                hidden_size=hidden_size,
                use_layer_norm=use_layer_norm,
                dropout=dropout,
            ),
            *[
                _ResidualBlock(
                    input_dim=hidden_size,
                    output_dim=hidden_size,
                    hidden_size=hidden_size,
                    use_layer_norm=use_layer_norm,
                    dropout=dropout,
                )
                for _ in range(num_encoder_layers - 1)
            ],
        )

        self.decoders = nn.Sequential(
            *[
                _ResidualBlock(
                    input_dim=hidden_size,
                    output_dim=hidden_size,
                    hidden_size=hidden_size,
                    use_layer_norm=use_layer_norm,
                    dropout=dropout,
                )
                for _ in range(num_decoder_layers - 1)
            ],
            # add decoder output layer
            _ResidualBlock(
                input_dim=hidden_size,
                output_dim=decoder_output_dim
                * self.output_chunk_length
                * self.nr_params,
                hidden_size=hidden_size,
                use_layer_norm=use_layer_norm,
                dropout=dropout,
            ),
        )

        decoder_input_dim = decoder_output_dim * self.nr_params
        if temporal_width_future and future_cov_dim:
            decoder_input_dim += temporal_width_future
        elif future_cov_dim:
            decoder_input_dim += future_cov_dim

        self.temporal_decoder = _ResidualBlock(
            input_dim=decoder_input_dim,
            output_dim=output_dim * self.nr_params,
            hidden_size=temporal_decoder_hidden,
            use_layer_norm=use_layer_norm,
            dropout=dropout,
        )

        self.lookback_skip = nn.Linear(
            self.input_chunk_length, self.output_chunk_length * self.nr_params
        )

    @io_processor
    def forward(
        self, x_in: tuple[torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]
    ) -> torch.Tensor:
        """TiDE model forward pass.
        Parameters
        ----------
        x_in
            comes as tuple `(x_past, x_future, x_static)` where `x_past` is the input/past chunk and `x_future`
            is the output/future chunk. Input dimensions are `(batch_size, time_steps, components)`
        Returns
        -------
        torch.Tensor
            The output Tensor of shape `(batch_size, output_chunk_length, output_dim, nr_params)`
        """

        # x has shape (batch_size, input_chunk_length, input_dim)
        # x_future_covariates has shape (batch_size, input_chunk_length, future_cov_dim)
        # x_static_covariates has shape (batch_size, static_cov_dim)
        x, x_future_covariates, x_static_covariates = x_in

        x_lookback = x[:, :, : self.output_dim]

        # future covariates: feature projection or raw features
        # historical future covariates need to be extracted from x and stacked with part of future covariates
        if self.future_cov_dim:
            x_dynamic_future_covariates = torch.cat(
                [
                    x[
                        :,
                        :,
                        None if self.future_cov_dim == 0 else -self.future_cov_dim :,
                    ],
                    x_future_covariates,
                ],
                dim=1,
            )
            if self.temporal_width_future:
                # project input features across all input and output time steps
                x_dynamic_future_covariates = self.future_cov_projection(
                    x_dynamic_future_covariates
                )
        else:
            x_dynamic_future_covariates = None

        # past covariates: feature projection or raw features
        # the past covariates are embedded in `x`
        if self.past_cov_dim:
            x_dynamic_past_covariates = x[
                :,
                :,
                self.output_dim : self.output_dim + self.past_cov_dim,
            ]
            if self.temporal_width_past:
                # project input features across all input time steps
                x_dynamic_past_covariates = self.past_cov_projection(
                    x_dynamic_past_covariates
                )
        else:
            x_dynamic_past_covariates = None

        # setup input to encoder
        encoded = [
            x_lookback,
            x_dynamic_past_covariates,
            x_dynamic_future_covariates,
            x_static_covariates,
        ]
        encoded = [t.flatten(start_dim=1) for t in encoded if t is not None]
        encoded = torch.cat(encoded, dim=1)

        # encoder, decode, reshape
        encoded = self.encoders(encoded)
        decoded = self.decoders(encoded)

        # get view that is batch size x output chunk length x self.decoder_output_dim x nr params
        decoded = decoded.view(x.shape[0], self.output_chunk_length, -1)

        # stack and temporally decode with future covariate last output steps
        temporal_decoder_input = [
            decoded,
            (
                x_dynamic_future_covariates[:, -self.output_chunk_length :, :]
                if self.future_cov_dim > 0
                else None
            ),
        ]
        temporal_decoder_input = [t for t in temporal_decoder_input if t is not None]

        temporal_decoder_input = torch.cat(temporal_decoder_input, dim=2)
        temporal_decoded = self.temporal_decoder(temporal_decoder_input)

        # pass x_lookback through self.lookback_skip but swap the last two dimensions
        # this is needed because the skip connection is applied across the input time steps
        # and not across the output time steps
        skip = self.lookback_skip(x_lookback.transpose(1, 2)).transpose(1, 2)

        # add skip connection
        y = temporal_decoded + skip.reshape_as(
            temporal_decoded
        )  # skip.view(temporal_decoded.shape)

        y = y.view(-1, self.output_chunk_length, self.output_dim, self.nr_params)
        return y


class TiDEModel(MixedCovariatesTorchModel):
    def __init__(
        self,
        input_chunk_length: int,
        output_chunk_length: int,
        output_chunk_shift: int = 0,
        num_encoder_layers: int = 1,
        num_decoder_layers: int = 1,
        decoder_output_dim: int = 16,
        hidden_size: int = 128,
        temporal_width_past: int = 4,
        temporal_width_future: int = 4,
        temporal_hidden_size_past: int = None,
        temporal_hidden_size_future: int = None,
        temporal_decoder_hidden: int = 32,
        use_layer_norm: bool = False,
        dropout: float = 0.1,
        use_static_covariates: bool = True,
        **kwargs,
    ):
        """An implementation of the TiDE model, as presented in [1]_.

        TiDE is similar to Transformers (implemented in :class:`TransformerModel`),
        but attempts to provide better performance at lower computational cost by introducing
        multilayer perceptron (MLP)-based encoder-decoders without attention.

        This model supports past covariates (known for `input_chunk_length` points before prediction time),
        future covariates (known for `output_chunk_length` points after prediction time), static covariates,
        as well as probabilistic forecasting.

        The encoder and decoder are implemented as a series of residual blocks. The number of residual blocks in
        the encoder and decoder can be controlled via ``num_encoder_layers`` and ``num_decoder_layers`` respectively.
        The width of the layers in the residual blocks can be controlled via ``hidden_size``. Similarly, the width
        of the layers in the temporal decoder can be controlled via ``temporal_decoder_hidden``.

        Parameters
        ----------
        input_chunk_length
            Number of time steps in the past to take as a model input (per chunk). Applies to the target
            series, and past and/or future covariates (if the model supports it).
        output_chunk_length
            Number of time steps predicted at once (per chunk) by the internal model. Also, the number of future values
            from future covariates to use as a model input (if the model supports future covariates). It is not the same
            as forecast horizon `n` used in `predict()`, which is the desired number of prediction points generated
            using either a one-shot- or autoregressive forecast. Setting `n <= output_chunk_length` prevents
            auto-regression. This is useful when the covariates don't extend far enough into the future, or to prohibit
            the model from using future values of past and / or future covariates for prediction (depending on the
            model's covariate support).
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future (relative to the
            input chunk end). This will create a gap between the input and output. If the model supports
            `future_covariates`, the future values are extracted from the shifted output chunk. Predictions will start
            `output_chunk_shift` steps after the end of the target `series`. If `output_chunk_shift` is set, the model
            cannot generate autoregressive predictions (`n > output_chunk_length`).
        num_encoder_layers
            The number of residual blocks in the encoder.
        num_decoder_layers
            The number of residual blocks in the decoder.
        decoder_output_dim
            The dimensionality of the output of the decoder.
        hidden_size
            The width of the layers in the residual blocks of the encoder and decoder.
        temporal_width_past
            The width of the output layer in the past covariate projection residual block. If `0`,
            will bypass feature projection and use the raw feature data.
        temporal_width_future
            The width of the output layer in the future covariate projection residual block. If `0`,
            will bypass feature projection and use the raw feature data.
        temporal_hidden_size_past
            The width of the hidden layer in the past covariate projection residual block. If not specified,
            defaults to `hidden_size`, which is the width of the hidden layer in the encoder and decoder.
            This is likely to be too large in many cases, so it is recommended to set this parameter explicitly.
        temporal_hidden_size_future
            The width of the hidden layer in the future covariate projection residual block. If not specified,
            defaults to `hidden_size`, which is the width of the hidden layer in the encoder and decoder.
            This is likely to be too large in many cases, so it is recommended to set this parameter explicitly.
        temporal_decoder_hidden
            The width of the layers in the temporal decoder.
        use_layer_norm
            Whether to use layer normalization in the residual blocks.
        dropout
            The dropout probability to be used in fully connected layers. This is compatible with Monte Carlo dropout
            at inference time for model uncertainty estimation (enabled with ``mc_dropout=True`` at
            prediction time).
        **kwargs
            Optional arguments to initialize the pytorch_lightning.Module, pytorch_lightning.Trainer, and
            Darts' :class:`TorchForecastingModel`.

        loss_fn
            PyTorch loss function used for training.
            This parameter will be ignored for probabilistic models if the ``likelihood`` parameter is specified.
            Default: ``torch.nn.MSELoss()``.
        likelihood
            One of Darts' :meth:`Likelihood <darts.utils.likelihood_models.Likelihood>` models to be used for
            probabilistic forecasts. Default: ``None``.
        torch_metrics
            A torch metric or a ``MetricCollection`` used for evaluation. A full list of available metrics can be found
            at https://torchmetrics.readthedocs.io/en/latest/. Default: ``None``.
        optimizer_cls
            The PyTorch optimizer class to be used. Default: ``torch.optim.Adam``.
        optimizer_kwargs
            Optionally, some keyword arguments for the PyTorch optimizer (e.g., ``{'lr': 1e-3}``
            for specifying a learning rate). Otherwise, the default values of the selected ``optimizer_cls``
            will be used. Default: ``None``.
        lr_scheduler_cls
            Optionally, the PyTorch learning rate scheduler class to be used. Specifying ``None`` corresponds
            to using a constant learning rate. Default: ``None``.
        lr_scheduler_kwargs
            Optionally, some keyword arguments for the PyTorch learning rate scheduler. Default: ``None``.
        use_reversible_instance_norm
            Whether to use reversible instance normalization `RINorm` against distribution shift as shown in [2]_.
            It is only applied to the features of the target series and not the covariates.
        batch_size
            Number of time series (input and output sequences) used in each training pass. Default: ``32``.
        n_epochs
            Number of epochs over which to train the model. Default: ``100``.
        model_name
            Name of the model. Used for creating checkpoints and saving tensorboard data. If not specified,
            defaults to the following string ``"YYYY-mm-dd_HH_MM_SS_torch_model_run_PID"``, where the initial part
            of the name is formatted with the local date and time, while PID is the processed ID (preventing models
            spawned at the same time by different processes to share the same model_name). E.g.,
            ``"2021-06-14_09_53_32_torch_model_run_44607"``.
        work_dir
            Path of the working directory, where to save checkpoints and Tensorboard summaries.
            Default: current working directory.
        log_tensorboard
            If set, use Tensorboard to log the different parameters. The logs will be located in:
            ``"{work_dir}/darts_logs/{model_name}/logs/"``. Default: ``False``.
        nr_epochs_val_period
            Number of epochs to wait before evaluating the validation loss (if a validation
            ``TimeSeries`` is passed to the :func:`fit()` method). Default: ``1``.
        force_reset
            If set to ``True``, any previously-existing model with the same name will be reset (all checkpoints will
            be discarded). Default: ``False``.
        save_checkpoints
            Whether to automatically save the untrained model and checkpoints from training.
            To load the model from checkpoint, call :func:`MyModelClass.load_from_checkpoint()`, where
            :class:`MyModelClass` is the :class:`TorchForecastingModel` class that was used (such as :class:`TFTModel`,
            :class:`NBEATSModel`, etc.). If set to ``False``, the model can still be manually saved using
            :func:`save()` and loaded using :func:`load()`. Default: ``False``.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        random_state
            Control the randomness of the weights initialization. Check this
            `link <https://scikit-learn.org/stable/glossary.html#term-random_state>`_ for more details.
            Default: ``None``.
        pl_trainer_kwargs
            By default :class:`TorchForecastingModel` creates a PyTorch Lightning Trainer with several useful presets
            that performs the training, validation and prediction processes. These presets include automatic
            checkpointing, tensorboard logging, setting the torch device and more.
            With ``pl_trainer_kwargs`` you can add additional kwargs to instantiate the PyTorch Lightning trainer
            object. Check the `PL Trainer documentation
            <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_ for more information about the
            supported kwargs. Default: ``None``.
            Running on GPU(s) is also possible using ``pl_trainer_kwargs`` by specifying keys ``"accelerator",
            "devices", and "auto_select_gpus"``. Some examples for setting the devices inside the ``pl_trainer_kwargs``
            dict:

            - ``{"accelerator": "cpu"}`` for CPU,
            - ``{"accelerator": "gpu", "devices": [i]}`` to use only GPU ``i`` (``i`` must be an integer),
            - ``{"accelerator": "gpu", "devices": -1, "auto_select_gpus": True}`` to use all available GPUS.

            For more info, see here:
            https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html#trainer-flags , and
            https://pytorch-lightning.readthedocs.io/en/stable/accelerators/gpu_basic.html#train-on-multiple-gpus

            With parameter ``"callbacks"`` you can add custom or PyTorch-Lightning built-in callbacks to Darts'
            :class:`TorchForecastingModel`. Below is an example for adding EarlyStopping to the training process.
            The model will stop training early if the validation loss `val_loss` does not improve beyond
            specifications. For more information on callbacks, visit:
            `PyTorch Lightning Callbacks
            <https://pytorch-lightning.readthedocs.io/en/stable/extensions/callbacks.html>`_

            .. highlight:: python
            .. code-block:: python

                from pytorch_lightning.callbacks.early_stopping import EarlyStopping

                # stop training when validation loss does not decrease more than 0.05 (`min_delta`) over
                # a period of 5 epochs (`patience`)
                my_stopper = EarlyStopping(
                    monitor="val_loss",
                    patience=5,
                    min_delta=0.05,
                    mode='min',
                )

                pl_trainer_kwargs={"callbacks": [my_stopper]}
            ..

            Note that you can also use a custom PyTorch Lightning Trainer for training and prediction with optional
            parameter ``trainer`` in :func:`fit()` and :func:`predict()`.
        show_warnings
            whether to show warnings raised from PyTorch Lightning. Useful to detect potential issues of
            your forecasting use case. Default: ``False``.

        References
        ----------
        .. [1] A. Das et al. "Long-term Forecasting with TiDE: Time-series Dense Encoder",
                http://arxiv.org/abs/2304.08424
        .. [2] T. Kim et al. "Reversible Instance Normalization for Accurate Time-Series Forecasting against
                Distribution Shift", https://openreview.net/forum?id=cGDAkQo1C0p

        Examples
        --------
        >>> from darts.datasets import WeatherDataset
        >>> from darts.models import TiDEModel
        >>> series = WeatherDataset().load()
        >>> # predicting atmospheric pressure
        >>> target = series['p (mbar)'][:100]
        >>> # optionally, use past observed rainfall (pretending to be unknown beyond index 100)
        >>> past_cov = series['rain (mm)'][:100]
        >>> # optionally, use future temperatures (pretending this component is a forecast)
        >>> future_cov = series['T (degC)'][:106]
        >>> model = TiDEModel(
        >>>     input_chunk_length=6,
        >>>     output_chunk_length=6,
        >>>     n_epochs=20
        >>> )
        >>> model.fit(target, past_covariates=past_cov, future_covariates=future_cov)
        >>> pred = model.predict(6)
        >>> pred.values()
        array([[1008.1667634 ],
               [ 997.08337201],
               [1017.72035839],
               [1005.10790392],
               [ 998.90537286],
               [1005.91534452]])

        .. note::
            `TiDE example notebook <https://unit8co.github.io/darts/examples/18-TiDE-examples.html>`_ presents
            techniques that can be used to improve the forecasts quality compared to this simple usage example.
        """
        if temporal_width_past < 0 or temporal_width_future < 0:
            raise_log(
                ValueError(
                    "`temporal_width_past` and `temporal_width_future` must be >= 0."
                ),
                logger=logger,
            )
        super().__init__(**self._extract_torch_model_params(**self.model_params))

        # extract pytorch lightning module kwargs
        self.pl_module_params = self._extract_pl_module_params(**self.model_params)

        self.num_encoder_layers = num_encoder_layers
        self.num_decoder_layers = num_decoder_layers
        self.decoder_output_dim = decoder_output_dim
        self.hidden_size = hidden_size
        self.temporal_width_past = temporal_width_past
        self.temporal_width_future = temporal_width_future
        self.temporal_hidden_size_past = temporal_hidden_size_past or hidden_size
        self.temporal_hidden_size_future = temporal_hidden_size_future or hidden_size
        self.temporal_decoder_hidden = temporal_decoder_hidden

        self._considers_static_covariates = use_static_covariates

        self.use_layer_norm = use_layer_norm
        self.dropout = dropout

    def _create_model(
        self, train_sample: MixedCovariatesTrainTensorType
    ) -> torch.nn.Module:
        (
            past_target,
            past_covariates,
            historic_future_covariates,
            future_covariates,
            static_covariates,
            future_target,
        ) = train_sample

        # target, past covariates, historic future covariates
        input_dim = (
            past_target.shape[1]
            + (past_covariates.shape[1] if past_covariates is not None else 0)
            + (
                historic_future_covariates.shape[1]
                if historic_future_covariates is not None
                else 0
            )
        )

        output_dim = future_target.shape[1]

        future_cov_dim = (
            future_covariates.shape[1] if future_covariates is not None else 0
        )
        static_cov_dim = (
            static_covariates.shape[0] * static_covariates.shape[1]
            if static_covariates is not None
            else 0
        )

        nr_params = 1 if self.likelihood is None else self.likelihood.num_parameters

        past_cov_dim = input_dim - output_dim - future_cov_dim
        if past_cov_dim and self.temporal_width_past >= past_cov_dim:
            logger.warning(
                f"number of `past_covariates` features is <= `temporal_width_past`, leading to feature expansion."
                f"number of covariates: {past_cov_dim}, `temporal_width_past={self.temporal_width_past}`."
            )
        if future_cov_dim and self.temporal_width_future >= future_cov_dim:
            logger.warning(
                f"number of `future_covariates` features is <= `temporal_width_future`, leading to feature expansion."
                f"number of covariates: {future_cov_dim}, `temporal_width_future={self.temporal_width_future}`."
            )

        return _TideModule(
            input_dim=input_dim,
            output_dim=output_dim,
            future_cov_dim=future_cov_dim,
            static_cov_dim=static_cov_dim,
            nr_params=nr_params,
            num_encoder_layers=self.num_encoder_layers,
            num_decoder_layers=self.num_decoder_layers,
            decoder_output_dim=self.decoder_output_dim,
            hidden_size=self.hidden_size,
            temporal_width_past=self.temporal_width_past,
            temporal_width_future=self.temporal_width_future,
            temporal_hidden_size_past=self.temporal_hidden_size_past,
            temporal_hidden_size_future=self.temporal_hidden_size_future,
            temporal_decoder_hidden=self.temporal_decoder_hidden,
            use_layer_norm=self.use_layer_norm,
            dropout=self.dropout,
            **self.pl_module_params,
        )

    @property
    def supports_static_covariates(self) -> bool:
        return True

    @property
    def supports_multivariate(self) -> bool:
        return True

    def _check_ckpt_parameters(self, tfm_save):
        # new parameters were added that will break loading weights
        new_params = ["temporal_hidden_size_past", "temporal_hidden_size_future"]
        for param in new_params:
            if param not in tfm_save.model_params:
                tfm_save.model_params[param] = None
        super()._check_ckpt_parameters(tfm_save)
